import six
from sqlalchemy import types

from ..exceptions import ImproperlyConfigured
from .scalar_coercible import ScalarCoercible


class TimezoneType(ScalarCoercible, types.TypeDecorator):
    """
    TimezoneType provides a way for saving timezones objects into database.
    TimezoneType saves timezone objects as strings on the way in and converts
    them back to objects when querying the database.

    ::

        from sqlalchemy_utils import TimezoneType

        class User(Base):
            __tablename__ = 'user'

            # Pass backend='pytz' to change it to use pytz. Other values:
            # 'dateutil' (default), and 'zoneinfo'.
            timezone = sa.Column(TimezoneType(backend='pytz'))

    :param backend: Whether to use 'dateutil', 'pytz' or 'zoneinfo' for
        timezones. 'zoneinfo' uses the standard library module in Python 3.9+,
        but requires the external 'backports.zoneinfo' package for older
        Python versions.

    """

    impl = types.Unicode(50)
    python_type = None
    cache_ok = True

    def __init__(self, backend='dateutil'):
        self.backend = backend
        if backend == 'dateutil':
            try:
                from dateutil.tz import tzfile
                from dateutil.zoneinfo import get_zonefile_instance

                self.python_type = tzfile
                self._to = get_zonefile_instance().zones.get
                self._from = lambda x: six.text_type(x._filename)

            except ImportError:
                raise ImproperlyConfigured(
                    "'python-dateutil' is required to use the "
                    "'dateutil' backend for 'TimezoneType'"
                )

        elif backend == 'pytz':
            try:
                from pytz import timezone
                from pytz.tzinfo import BaseTzInfo

                self.python_type = BaseTzInfo
                self._to = timezone
                self._from = six.text_type

            except ImportError:
                raise ImproperlyConfigured(
                    "'pytz' is required to use the 'pytz' backend "
                    "for 'TimezoneType'"
                )

        elif backend == "zoneinfo":
            try:
                import zoneinfo
            except ImportError:
                try:
                    from backports import zoneinfo
                except ImportError:
                    raise ImproperlyConfigured(
                        "'backports.zoneinfo' is required to use "
                        "the 'zoneinfo' backend for 'TimezoneType'"
                        "on Python version < 3.9"
                    )

            self.python_type = zoneinfo.ZoneInfo
            self._to = zoneinfo.ZoneInfo
            self._from = six.text_type

        else:
            raise ImproperlyConfigured(
                "'pytz', 'dateutil' or 'zoneinfo' are the backends "
                "supported for 'TimezoneType'"
            )

    def _coerce(self, value):
        if value is not None and not isinstance(value, self.python_type):
            obj = self._to(value)
            if obj is None:
                raise ValueError("unknown time zone '%s'" % value)
            return obj
        return value

    def process_bind_param(self, value, dialect):
        return self._from(self._coerce(value)) if value else None

    def process_result_value(self, value, dialect):
        return self._to(value) if value else None



class ScalarCoercible:
    cache_ok = True

    def _coerce(self, value):
        raise NotImplementedError

    def coercion_listener(self, target, value, oldvalue, initiator):
        return self._coerce(value)



#!/usr/bin/env python

import os

__author__ = "Daniel Grossmann-Kavanagh"
__version__ = '0.0.9'


def get_path(rel):
    return os.path.join(
        os.path.abspath(os.path.dirname(os.path.realpath(__file__))), rel
    )


swagger_ui_2_2_10_path = get_path("vendor/swagger-ui-2.2.10")

swagger_ui_3_52_0_path = get_path("vendor/swagger-ui-3.52.0")

# latest major versions
swagger_ui_2_path = swagger_ui_2_2_10_path
swagger_ui_3_path = swagger_ui_3_52_0_path

# default to swagger 3
swagger_ui_path = swagger_ui_3_path



import datetime
from warnings import warn

from jwt import (
    ExpiredSignatureError, InvalidTokenError, InvalidAudienceError,
    InvalidIssuerError, DecodeError
)

try:
    from flask import _app_ctx_stack as ctx_stack
except ImportError:  # pragma: no cover
    from flask import _request_ctx_stack as ctx_stack

from flask_jwt_extended.config import config
from flask_jwt_extended.exceptions import (
    JWTDecodeError, NoAuthorizationError, InvalidHeaderError, WrongTokenError,
    RevokedTokenError, FreshTokenRequired, CSRFError, UserLoadError,
    UserClaimsVerificationError
)
from flask_jwt_extended.default_callbacks import (
    default_expired_token_callback, default_user_claims_callback,
    default_user_identity_callback, default_invalid_token_callback,
    default_unauthorized_callback, default_needs_fresh_token_callback,
    default_revoked_token_callback, default_user_loader_error_callback,
    default_claims_verification_callback, default_verify_claims_failed_callback,
    default_decode_key_callback, default_encode_key_callback,
    default_jwt_headers_callback)
from flask_jwt_extended.tokens import (
    encode_refresh_token, encode_access_token
)
from flask_jwt_extended.utils import get_jwt_identity


class JWTManager(object):
    """
    An object used to hold JWT settings and callback functions for the
    Flask-JWT-Extended extension.

    Instances of :class:`JWTManager` are *not* bound to specific apps, so
    you can create one in the main body of your code and then bind it
    to your app in a factory function.
    """

    def __init__(self, app=None):
        """
        Create the JWTManager instance. You can either pass a flask application
        in directly here to register this extension with the flask app, or
        call init_app after creating this object (in a factory pattern).

        :param app: A flask application
        """
        # Register the default error handler callback methods. These can be
        # overridden with the appropriate loader decorators
        self._user_claims_callback = default_user_claims_callback
        self._user_identity_callback = default_user_identity_callback
        self._expired_token_callback = default_expired_token_callback
        self._invalid_token_callback = default_invalid_token_callback
        self._unauthorized_callback = default_unauthorized_callback
        self._needs_fresh_token_callback = default_needs_fresh_token_callback
        self._revoked_token_callback = default_revoked_token_callback
        self._user_loader_callback = None
        self._user_loader_error_callback = default_user_loader_error_callback
        self._token_in_blacklist_callback = None
        self._claims_verification_callback = default_claims_verification_callback
        self._verify_claims_failed_callback = default_verify_claims_failed_callback
        self._decode_key_callback = default_decode_key_callback
        self._encode_key_callback = default_encode_key_callback
        self._jwt_additional_header_callback = default_jwt_headers_callback

        # Register this extension with the flask app now (if it is provided)
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """
        Register this extension with the flask app.

        :param app: A flask application
        """
        # Save this so we can use it later in the extension
        if not hasattr(app, 'extensions'):   # pragma: no cover
            app.extensions = {}
        app.extensions['flask-jwt-extended'] = self

        # Set all the default configurations for this extension
        self._set_default_configuration_options(app)
        self._set_error_handler_callbacks(app)

    def _set_error_handler_callbacks(self, app):
        """
        Sets the error handler callbacks used by this extension
        """
        @app.errorhandler(NoAuthorizationError)
        def handle_auth_error(e):
            return self._unauthorized_callback(str(e))

        @app.errorhandler(CSRFError)
        def handle_csrf_error(e):
            return self._unauthorized_callback(str(e))

        @app.errorhandler(ExpiredSignatureError)
        def handle_expired_error(e):
            try:
                token = ctx_stack.top.expired_jwt
                return self._expired_token_callback(token)
            except TypeError:
                msg = (
                    "jwt.expired_token_loader callback now takes the expired token "
                    "as an additional parameter. Example: expired_callback(token)"
                )
                warn(msg, DeprecationWarning)
                return self._expired_token_callback()

        @app.errorhandler(InvalidHeaderError)
        def handle_invalid_header_error(e):
            return self._invalid_token_callback(str(e))

        @app.errorhandler(DecodeError)
        def handle_invalid_header_error(e):
            return self._invalid_token_callback(str(e))

        @app.errorhandler(InvalidTokenError)
        def handle_invalid_token_error(e):
            return self._invalid_token_callback(str(e))

        @app.errorhandler(JWTDecodeError)
        def handle_jwt_decode_error(e):
            return self._invalid_token_callback(str(e))

        @app.errorhandler(WrongTokenError)
        def handle_wrong_token_error(e):
            return self._invalid_token_callback(str(e))

        @app.errorhandler(InvalidAudienceError)
        def handle_invalid_audience_error(e):
            return self._invalid_token_callback(str(e))

        @app.errorhandler(InvalidIssuerError)
        def handle_invalid_issuer_error(e):
            return self._invalid_token_callback(str(e))

        @app.errorhandler(RevokedTokenError)
        def handle_revoked_token_error(e):
            return self._revoked_token_callback()

        @app.errorhandler(FreshTokenRequired)
        def handle_fresh_token_required(e):
            return self._needs_fresh_token_callback()

        @app.errorhandler(UserLoadError)
        def handler_user_load_error(e):
            # The identity is already saved before this exception was raised,
            # otherwise a different exception would be raised, which is why we
            # can safely call get_jwt_identity() here
            identity = get_jwt_identity()
            return self._user_loader_error_callback(identity)

        @app.errorhandler(UserClaimsVerificationError)
        def handle_failed_user_claims_verification(e):
            return self._verify_claims_failed_callback()

    @staticmethod
    def _set_default_configuration_options(app):
        """
        Sets the default configuration options used by this extension
        """
        # Where to look for the JWT. Available options are cookies or headers
        app.config.setdefault('JWT_TOKEN_LOCATION', ('headers',))

        # Options for JWTs when the TOKEN_LOCATION is headers
        app.config.setdefault('JWT_HEADER_NAME', 'Authorization')
        app.config.setdefault('JWT_HEADER_TYPE', 'Bearer')

        # Options for JWTs then the TOKEN_LOCATION is query_string
        app.config.setdefault('JWT_QUERY_STRING_NAME', 'jwt')

        # Option for JWTs when the TOKEN_LOCATION is cookies
        app.config.setdefault('JWT_ACCESS_COOKIE_NAME', 'access_token_cookie')
        app.config.setdefault('JWT_REFRESH_COOKIE_NAME', 'refresh_token_cookie')
        app.config.setdefault('JWT_ACCESS_COOKIE_PATH', '/')
        app.config.setdefault('JWT_REFRESH_COOKIE_PATH', '/')
        app.config.setdefault('JWT_COOKIE_SECURE', False)
        app.config.setdefault('JWT_COOKIE_DOMAIN', None)
        app.config.setdefault('JWT_SESSION_COOKIE', True)
        app.config.setdefault('JWT_COOKIE_SAMESITE', None)

        # Option for JWTs when the TOKEN_LOCATION is json
        app.config.setdefault('JWT_JSON_KEY', 'access_token')
        app.config.setdefault('JWT_REFRESH_JSON_KEY', 'refresh_token')

        # Options for using double submit csrf protection
        app.config.setdefault('JWT_COOKIE_CSRF_PROTECT', True)
        app.config.setdefault('JWT_CSRF_METHODS', ['POST', 'PUT', 'PATCH', 'DELETE'])
        app.config.setdefault('JWT_ACCESS_CSRF_HEADER_NAME', 'X-CSRF-TOKEN')
        app.config.setdefault('JWT_REFRESH_CSRF_HEADER_NAME', 'X-CSRF-TOKEN')
        app.config.setdefault('JWT_CSRF_IN_COOKIES', True)
        app.config.setdefault('JWT_ACCESS_CSRF_COOKIE_NAME', 'csrf_access_token')
        app.config.setdefault('JWT_REFRESH_CSRF_COOKIE_NAME', 'csrf_refresh_token')
        app.config.setdefault('JWT_ACCESS_CSRF_COOKIE_PATH', '/')
        app.config.setdefault('JWT_REFRESH_CSRF_COOKIE_PATH', '/')
        app.config.setdefault('JWT_CSRF_CHECK_FORM', False)
        app.config.setdefault('JWT_ACCESS_CSRF_FIELD_NAME', 'csrf_token')
        app.config.setdefault('JWT_REFRESH_CSRF_FIELD_NAME', 'csrf_token')

        # How long an a token will live before they expire.
        app.config.setdefault('JWT_ACCESS_TOKEN_EXPIRES', datetime.timedelta(minutes=15))
        app.config.setdefault('JWT_REFRESH_TOKEN_EXPIRES', datetime.timedelta(days=30))

        # What algorithm to use to sign the token. See here for a list of options:
        # https://github.com/jpadilla/pyjwt/blob/master/jwt/api_jwt.py
        app.config.setdefault('JWT_ALGORITHM', 'HS256')

        # What algorithms are allowed to decode a token
        app.config.setdefault('JWT_DECODE_ALGORITHMS', None)

        # Secret key to sign JWTs with. Only used if a symmetric algorithm is
        # used (such as the HS* algorithms). We will use the app secret key
        # if this is not set.
        app.config.setdefault('JWT_SECRET_KEY', None)

        # Keys to sign JWTs with when use when using an asymmetric
        # (public/private key) algorithm, such as RS* or EC*
        app.config.setdefault('JWT_PRIVATE_KEY', None)
        app.config.setdefault('JWT_PUBLIC_KEY', None)

        # Options for blacklisting/revoking tokens
        app.config.setdefault('JWT_BLACKLIST_ENABLED', False)
        app.config.setdefault('JWT_BLACKLIST_TOKEN_CHECKS', ('access', 'refresh'))

        app.config.setdefault('JWT_IDENTITY_CLAIM', 'identity')
        app.config.setdefault('JWT_USER_CLAIMS', 'user_claims')
        app.config.setdefault('JWT_DECODE_AUDIENCE', None)
        app.config.setdefault('JWT_ENCODE_ISSUER', None)
        app.config.setdefault('JWT_DECODE_ISSUER', None)
        app.config.setdefault('JWT_DECODE_LEEWAY', 0)

        app.config.setdefault('JWT_CLAIMS_IN_REFRESH_TOKEN', False)

        app.config.setdefault('JWT_ERROR_MESSAGE_KEY', 'msg')

    def user_claims_loader(self, callback):
        """
        This decorator sets the callback function for adding custom claims to an
        access token when :func:`~flask_jwt_extended.create_access_token` is
        called. By default, no extra user claims will be added to the JWT.

        *HINT*: The callback function must be a function that takes only **one** argument,
        which is the object passed into
        :func:`~flask_jwt_extended.create_access_token`, and returns the custom
        claims you want included in the access tokens. This returned claims
        must be *JSON serializable*.
        """
        self._user_claims_callback = callback
        return callback

    def user_identity_loader(self, callback):
        """
        This decorator sets the callback function for getting the JSON
        serializable identity out of whatever object is passed into
        :func:`~flask_jwt_extended.create_access_token` and
        :func:`~flask_jwt_extended.create_refresh_token`. By default, this will
        return the unmodified object that is passed in as the `identity` kwarg
        to the above functions.

        *HINT*: The callback function must be a function that takes only **one** argument,
        which is the object passed into
        :func:`~flask_jwt_extended.create_access_token` or
        :func:`~flask_jwt_extended.create_refresh_token`, and returns the
        *JSON serializable* identity of this token.
        """
        self._user_identity_callback = callback
        return callback

    def expired_token_loader(self, callback):
        """
        This decorator sets the callback function that will be called if an
        expired JWT attempts to access a protected endpoint. The default
        implementation will return a 401 status code with the JSON:

        {"msg": "Token has expired"}

        *HINT*: The callback must be a function that takes **one** argument,
        which is a dictionary containing the data for the expired token, and
        and returns a *Flask response*.
        """
        self._expired_token_callback = callback
        return callback

    def invalid_token_loader(self, callback):
        """
        This decorator sets the callback function that will be called if an
        invalid JWT attempts to access a protected endpoint. The default
        implementation will return a 422 status code with the JSON:

        {"msg": "<error description>"}

        *HINT*: The callback must be a function that takes only **one** argument, which is
        a string which contains the reason why a token is invalid, and returns
        a *Flask response*.
        """
        self._invalid_token_callback = callback
        return callback

    def unauthorized_loader(self, callback):
        """
        This decorator sets the callback function that will be called if an
        no JWT can be found when attempting to access a protected endpoint.
        The default implementation will return a 401 status code with the JSON:

        {"msg": "<error description>"}

        *HINT*: The callback must be a function that takes only **one** argument, which is
        a string which contains the reason why a JWT could not be found, and
        returns a *Flask response*.
        """
        self._unauthorized_callback = callback
        return callback

    def needs_fresh_token_loader(self, callback):
        """
        This decorator sets the callback function that will be called if a
        valid and non-fresh token attempts to access an endpoint protected with
        the :func:`~flask_jwt_extended.fresh_jwt_required` decorator. The
        default implementation will return a 401 status code with the JSON:

        {"msg": "Fresh token required"}

        *HINT*: The callback must be a function that takes **no** arguments, and returns
        a *Flask response*.
        """
        self._needs_fresh_token_callback = callback
        return callback

    def revoked_token_loader(self, callback):
        """
        This decorator sets the callback function that will be called if a
        revoked token attempts to access a protected endpoint. The default
        implementation will return a 401 status code with the JSON:

        {"msg": "Token has been revoked"}

        *HINT*: The callback must be a function that takes **no** arguments, and returns
        a *Flask response*.
        """
        self._revoked_token_callback = callback
        return callback

    def user_loader_callback_loader(self, callback):
        """
        This decorator sets the callback function that will be called to
        automatically load an object when a protected endpoint is accessed.
        By default this is not used.

        *HINT*: The callback must take **one** argument which is the identity JWT
        accessing the protected endpoint, and it must return any object (which can
        then be accessed via the :attr:`~flask_jwt_extended.current_user` LocalProxy
        in the protected endpoint), or `None` in the case of a user not being
        able to be loaded for any reason. If this callback function returns
        `None`, the :meth:`~flask_jwt_extended.JWTManager.user_loader_error_loader`
        will be called.
        """
        self._user_loader_callback = callback
        return callback

    def user_loader_error_loader(self, callback):
        """
        This decorator sets the callback function that will be called if `None`
        is returned from the
        :meth:`~flask_jwt_extended.JWTManager.user_loader_callback_loader`
        callback function. The default implementation will return
        a 401 status code with the JSON:

        {"msg": "Error loading the user <identity>"}

        *HINT*: The callback must be a function that takes **one** argument, which is the
        identity of the user who failed to load, and must return a *Flask response*.
        """
        self._user_loader_error_callback = callback
        return callback

    def token_in_blacklist_loader(self, callback):
        """
        This decorator sets the callback function that will be called when
        a protected endpoint is accessed and will check if the JWT has been
        been revoked. By default, this callback is not used.

        *HINT*: The callback must be a function that takes **one** argument, which is the
        decoded JWT (python dictionary), and returns *`True`* if the token
        has been blacklisted (or is otherwise considered revoked), or *`False`*
        otherwise.
        """
        self._token_in_blacklist_callback = callback
        return callback

    def claims_verification_loader(self, callback):
        """
        This decorator sets the callback function that will be called when
        a protected endpoint is accessed, and will check if the custom claims
        in the JWT are valid. By default, this callback is not used. The
        error returned if the claims are invalid can be controlled via the
        :meth:`~flask_jwt_extended.JWTManager.claims_verification_failed_loader`
        decorator.

        *HINT*: This callback must be a function that takes **one** argument, which is the
        custom claims (python dict) present in the JWT, and returns *`True`* if the
        claims are valid, or *`False`* otherwise.
        """
        self._claims_verification_callback = callback
        return callback

    def claims_verification_failed_loader(self, callback):
        """
        This decorator sets the callback function that will be called if
        the :meth:`~flask_jwt_extended.JWTManager.claims_verification_loader`
        callback returns False, indicating that the user claims are not valid.
        The default implementation will return a 400 status code with the JSON:

        {"msg": "User claims verification failed"}

        *HINT*: This callback must be a function that takes **no** arguments, and returns
        a *Flask response*.
        """
        self._verify_claims_failed_callback = callback
        return callback

    def decode_key_loader(self, callback):
        """
        This decorator sets the callback function for getting the JWT decode key and
        can be used to dynamically choose the appropriate decode key based on token
        contents.

        The default implementation returns the decode key specified by
        `JWT_SECRET_KEY` or `JWT_PUBLIC_KEY`, depending on the signing algorithm.

        *HINT*: The callback function should be a function that takes
        **two** arguments, which are the unverified claims and headers of the jwt
        (dictionaries). The function must return a *string* which is the decode key
        in PEM format to verify the token.
        """
        self._decode_key_callback = callback
        return callback

    def encode_key_loader(self, callback):
        """
        This decorator sets the callback function for getting the JWT encode key and
        can be used to dynamically choose the appropriate encode key based on the
        token identity.

        The default implementation returns the encode key specified by
        `JWT_SECRET_KEY` or `JWT_PRIVATE_KEY`, depending on the signing algorithm.

        *HINT*: The callback function must be a function that takes only **one**
        argument, which is the identity as passed into the create_access_token
        or create_refresh_token functions, and must return a *string* which is
        the decode key to verify the token.
        """
        self._encode_key_callback = callback
        return callback

    def additional_headers_loader(self, callback):
        """
        This decorator sets the callback function for adding custom headers to an
        access token when :func:`~flask_jwt_extended.create_access_token` is
        called. By default, two headers will be added the type of the token, which is JWT,
        and the signing algorithm being used, such as HMAC SHA256 or RSA.

        *HINT*: The callback function must be a function that takes **no** argument,
        which is the object passed into
        :func:`~flask_jwt_extended.create_access_token`, and returns the custom
        claims you want included in the access tokens. This returned claims
        must be *JSON serializable*.
        """
        self._jwt_additional_header_callback = callback
        return callback

    def _create_refresh_token(self, identity, expires_delta=None, user_claims=None,
                              headers=None):
        if expires_delta is None:
            expires_delta = config.refresh_expires

        if user_claims is None and config.user_claims_in_refresh_token:
            user_claims = self._user_claims_callback(identity)

        if headers is None:
            headers = self._jwt_additional_header_callback(identity)

        refresh_token = encode_refresh_token(
            identity=self._user_identity_callback(identity),
            secret=self._encode_key_callback(identity),
            algorithm=config.algorithm,
            expires_delta=expires_delta,
            user_claims=user_claims,
            csrf=config.csrf_protect,
            identity_claim_key=config.identity_claim_key,
            user_claims_key=config.user_claims_key,
            json_encoder=config.json_encoder,
            headers=headers
        )
        return refresh_token

    def _create_access_token(self, identity, fresh=False, expires_delta=None,
                             user_claims=None, headers=None):
        if expires_delta is None:
            expires_delta = config.access_expires

        if user_claims is None:
            user_claims = self._user_claims_callback(identity)

        if headers is None:
            headers = self._jwt_additional_header_callback(identity)

        access_token = encode_access_token(
            identity=self._user_identity_callback(identity),
            secret=self._encode_key_callback(identity),
            algorithm=config.algorithm,
            expires_delta=expires_delta,
            fresh=fresh,
            user_claims=user_claims,
            csrf=config.csrf_protect,
            identity_claim_key=config.identity_claim_key,
            user_claims_key=config.user_claims_key,
            json_encoder=config.json_encoder,
            headers=headers,
            issuer=config.encode_issuer,
        )
        return access_token



from io import BytesIO
from os import PathLike
from subprocess import DEVNULL, PIPE, CalledProcessError, CompletedProcess
from typing import AsyncIterable, List, Mapping, Optional, Sequence, Union, cast

from ..abc import Process
from ._eventloop import get_asynclib
from ._tasks import create_task_group


async def run_process(command: Union[str, Sequence[str]], *, input: Optional[bytes] = None,
                      stdout: int = PIPE, stderr: int = PIPE, check: bool = True,
                      cwd: Union[str, bytes, 'PathLike[str]', None] = None,
                      env: Optional[Mapping[str, str]] = None, start_new_session: bool = False,
                      ) -> 'CompletedProcess[bytes]':
    """
    Run an external command in a subprocess and wait until it completes.

    .. seealso:: :func:`subprocess.run`

    :param command: either a string to pass to the shell, or an iterable of strings containing the
        executable name or path and its arguments
    :param input: bytes passed to the standard input of the subprocess
    :param stdout: either :data:`subprocess.PIPE` or :data:`subprocess.DEVNULL`
    :param stderr: one of :data:`subprocess.PIPE`, :data:`subprocess.DEVNULL` or
        :data:`subprocess.STDOUT`
    :param check: if ``True``, raise :exc:`~subprocess.CalledProcessError` if the process
        terminates with a return code other than 0
    :param cwd: If not ``None``, change the working directory to this before running the command
    :param env: if not ``None``, this mapping replaces the inherited environment variables from the
        parent process
    :param start_new_session: if ``true`` the setsid() system call will be made in the child
        process prior to the execution of the subprocess. (POSIX only)
    :return: an object representing the completed process
    :raises ~subprocess.CalledProcessError: if ``check`` is ``True`` and the process exits with a
        nonzero return code

    """
    async def drain_stream(stream: AsyncIterable[bytes], index: int) -> None:
        buffer = BytesIO()
        async for chunk in stream:
            buffer.write(chunk)

        stream_contents[index] = buffer.getvalue()

    async with await open_process(command, stdin=PIPE if input else DEVNULL, stdout=stdout,
                                  stderr=stderr, cwd=cwd, env=env,
                                  start_new_session=start_new_session) as process:
        stream_contents: List[Optional[bytes]] = [None, None]
        try:
            async with create_task_group() as tg:
                if process.stdout:
                    tg.start_soon(drain_stream, process.stdout, 0)
                if process.stderr:
                    tg.start_soon(drain_stream, process.stderr, 1)
                if process.stdin and input:
                    await process.stdin.send(input)
                    await process.stdin.aclose()

                await process.wait()
        except BaseException:
            process.kill()
            raise

    output, errors = stream_contents
    if check and process.returncode != 0:
        raise CalledProcessError(cast(int, process.returncode), command, output, errors)

    return CompletedProcess(command, cast(int, process.returncode), output, errors)


async def open_process(command: Union[str, Sequence[str]], *, stdin: int = PIPE,
                       stdout: int = PIPE, stderr: int = PIPE,
                       cwd: Union[str, bytes, 'PathLike[str]', None] = None,
                       env: Optional[Mapping[str, str]] = None,
                       start_new_session: bool = False) -> Process:
    """
    Start an external command in a subprocess.

    .. seealso:: :class:`subprocess.Popen`

    :param command: either a string to pass to the shell, or an iterable of strings containing the
        executable name or path and its arguments
    :param stdin: either :data:`subprocess.PIPE` or :data:`subprocess.DEVNULL`
    :param stdout: either :data:`subprocess.PIPE` or :data:`subprocess.DEVNULL`
    :param stderr: one of :data:`subprocess.PIPE`, :data:`subprocess.DEVNULL` or
        :data:`subprocess.STDOUT`
    :param cwd: If not ``None``, the working directory is changed before executing
    :param env: If env is not ``None``, it must be a mapping that defines the environment
        variables for the new process
    :param start_new_session: if ``true`` the setsid() system call will be made in the child
        process prior to the execution of the subprocess. (POSIX only)
    :return: an asynchronous process object

    """
    shell = isinstance(command, str)
    return await get_asynclib().open_process(command, shell=shell, stdin=stdin, stdout=stdout,
                                             stderr=stderr, cwd=cwd, env=env,
                                             start_new_session=start_new_session)



# -*- coding: utf-8 -*-

timezones = (
    "Africa/Abidjan",
    "Africa/Accra",
    "Africa/Addis_Ababa",
    "Africa/Algiers",
    "Africa/Asmara",
    "Africa/Asmera",
    "Africa/Bamako",
    "Africa/Bangui",
    "Africa/Banjul",
    "Africa/Bissau",
    "Africa/Blantyre",
    "Africa/Brazzaville",
    "Africa/Bujumbura",
    "Africa/Cairo",
    "Africa/Casablanca",
    "Africa/Ceuta",
    "Africa/Conakry",
    "Africa/Dakar",
    "Africa/Dar_es_Salaam",
    "Africa/Djibouti",
    "Africa/Douala",
    "Africa/El_Aaiun",
    "Africa/Freetown",
    "Africa/Gaborone",
    "Africa/Harare",
    "Africa/Johannesburg",
    "Africa/Juba",
    "Africa/Kampala",
    "Africa/Khartoum",
    "Africa/Kigali",
    "Africa/Kinshasa",
    "Africa/Lagos",
    "Africa/Libreville",
    "Africa/Lome",
    "Africa/Luanda",
    "Africa/Lubumbashi",
    "Africa/Lusaka",
    "Africa/Malabo",
    "Africa/Maputo",
    "Africa/Maseru",
    "Africa/Mbabane",
    "Africa/Mogadishu",
    "Africa/Monrovia",
    "Africa/Nairobi",
    "Africa/Ndjamena",
    "Africa/Niamey",
    "Africa/Nouakchott",
    "Africa/Ouagadougou",
    "Africa/Porto-Novo",
    "Africa/Sao_Tome",
    "Africa/Timbuktu",
    "Africa/Tripoli",
    "Africa/Tunis",
    "Africa/Windhoek",
    "America/Adak",
    "America/Anchorage",
    "America/Anguilla",
    "America/Antigua",
    "America/Araguaina",
    "America/Argentina/Buenos_Aires",
    "America/Argentina/Catamarca",
    "America/Argentina/ComodRivadavia",
    "America/Argentina/Cordoba",
    "America/Argentina/Jujuy",
    "America/Argentina/La_Rioja",
    "America/Argentina/Mendoza",
    "America/Argentina/Rio_Gallegos",
    "America/Argentina/Salta",
    "America/Argentina/San_Juan",
    "America/Argentina/San_Luis",
    "America/Argentina/Tucuman",
    "America/Argentina/Ushuaia",
    "America/Aruba",
    "America/Asuncion",
    "America/Atikokan",
    "America/Atka",
    "America/Bahia",
    "America/Bahia_Banderas",
    "America/Barbados",
    "America/Belem",
    "America/Belize",
    "America/Blanc-Sablon",
    "America/Boa_Vista",
    "America/Bogota",
    "America/Boise",
    "America/Buenos_Aires",
    "America/Cambridge_Bay",
    "America/Campo_Grande",
    "America/Cancun",
    "America/Caracas",
    "America/Catamarca",
    "America/Cayenne",
    "America/Cayman",
    "America/Chicago",
    "America/Chihuahua",
    "America/Coral_Harbour",
    "America/Cordoba",
    "America/Costa_Rica",
    "America/Creston",
    "America/Cuiaba",
    "America/Curacao",
    "America/Danmarkshavn",
    "America/Dawson",
    "America/Dawson_Creek",
    "America/Denver",
    "America/Detroit",
    "America/Dominica",
    "America/Edmonton",
    "America/Eirunepe",
    "America/El_Salvador",
    "America/Ensenada",
    "America/Fort_Nelson",
    "America/Fort_Wayne",
    "America/Fortaleza",
    "America/Glace_Bay",
    "America/Godthab",
    "America/Goose_Bay",
    "America/Grand_Turk",
    "America/Grenada",
    "America/Guadeloupe",
    "America/Guatemala",
    "America/Guayaquil",
    "America/Guyana",
    "America/Halifax",
    "America/Havana",
    "America/Hermosillo",
    "America/Indiana/Indianapolis",
    "America/Indiana/Knox",
    "America/Indiana/Marengo",
    "America/Indiana/Petersburg",
    "America/Indiana/Tell_City",
    "America/Indiana/Vevay",
    "America/Indiana/Vincennes",
    "America/Indiana/Winamac",
    "America/Indianapolis",
    "America/Inuvik",
    "America/Iqaluit",
    "America/Jamaica",
    "America/Jujuy",
    "America/Juneau",
    "America/Kentucky/Louisville",
    "America/Kentucky/Monticello",
    "America/Knox_IN",
    "America/Kralendijk",
    "America/La_Paz",
    "America/Lima",
    "America/Los_Angeles",
    "America/Louisville",
    "America/Lower_Princes",
    "America/Maceio",
    "America/Managua",
    "America/Manaus",
    "America/Marigot",
    "America/Martinique",
    "America/Matamoros",
    "America/Mazatlan",
    "America/Mendoza",
    "America/Menominee",
    "America/Merida",
    "America/Metlakatla",
    "America/Mexico_City",
    "America/Miquelon",
    "America/Moncton",
    "America/Monterrey",
    "America/Montevideo",
    "America/Montreal",
    "America/Montserrat",
    "America/Nassau",
    "America/New_York",
    "America/Nipigon",
    "America/Nome",
    "America/Noronha",
    "America/North_Dakota/Beulah",
    "America/North_Dakota/Center",
    "America/North_Dakota/New_Salem",
    "America/Nuuk",
    "America/Ojinaga",
    "America/Panama",
    "America/Pangnirtung",
    "America/Paramaribo",
    "America/Phoenix",
    "America/Port-au-Prince",
    "America/Port_of_Spain",
    "America/Porto_Acre",
    "America/Porto_Velho",
    "America/Puerto_Rico",
    "America/Punta_Arenas",
    "America/Rainy_River",
    "America/Rankin_Inlet",
    "America/Recife",
    "America/Regina",
    "America/Resolute",
    "America/Rio_Branco",
    "America/Rosario",
    "America/Santa_Isabel",
    "America/Santarem",
    "America/Santiago",
    "America/Santo_Domingo",
    "America/Sao_Paulo",
    "America/Scoresbysund",
    "America/Shiprock",
    "America/Sitka",
    "America/St_Barthelemy",
    "America/St_Johns",
    "America/St_Kitts",
    "America/St_Lucia",
    "America/St_Thomas",
    "America/St_Vincent",
    "America/Swift_Current",
    "America/Tegucigalpa",
    "America/Thule",
    "America/Thunder_Bay",
    "America/Tijuana",
    "America/Toronto",
    "America/Tortola",
    "America/Vancouver",
    "America/Virgin",
    "America/Whitehorse",
    "America/Winnipeg",
    "America/Yakutat",
    "America/Yellowknife",
    "Antarctica/Casey",
    "Antarctica/Davis",
    "Antarctica/DumontDUrville",
    "Antarctica/Macquarie",
    "Antarctica/Mawson",
    "Antarctica/McMurdo",
    "Antarctica/Palmer",
    "Antarctica/Rothera",
    "Antarctica/South_Pole",
    "Antarctica/Syowa",
    "Antarctica/Troll",
    "Antarctica/Vostok",
    "Arctic/Longyearbyen",
    "Asia/Aden",
    "Asia/Almaty",
    "Asia/Amman",
    "Asia/Anadyr",
    "Asia/Aqtau",
    "Asia/Aqtobe",
    "Asia/Ashgabat",
    "Asia/Ashkhabad",
    "Asia/Atyrau",
    "Asia/Baghdad",
    "Asia/Bahrain",
    "Asia/Baku",
    "Asia/Bangkok",
    "Asia/Barnaul",
    "Asia/Beirut",
    "Asia/Bishkek",
    "Asia/Brunei",
    "Asia/Calcutta",
    "Asia/Chita",
    "Asia/Choibalsan",
    "Asia/Chongqing",
    "Asia/Chungking",
    "Asia/Colombo",
    "Asia/Dacca",
    "Asia/Damascus",
    "Asia/Dhaka",
    "Asia/Dili",
    "Asia/Dubai",
    "Asia/Dushanbe",
    "Asia/Famagusta",
    "Asia/Gaza",
    "Asia/Harbin",
    "Asia/Hebron",
    "Asia/Ho_Chi_Minh",
    "Asia/Hong_Kong",
    "Asia/Hovd",
    "Asia/Irkutsk",
    "Asia/Istanbul",
    "Asia/Jakarta",
    "Asia/Jayapura",
    "Asia/Jerusalem",
    "Asia/Kabul",
    "Asia/Kamchatka",
    "Asia/Karachi",
    "Asia/Kashgar",
    "Asia/Kathmandu",
    "Asia/Katmandu",
    "Asia/Khandyga",
    "Asia/Kolkata",
    "Asia/Krasnoyarsk",
    "Asia/Kuala_Lumpur",
    "Asia/Kuching",
    "Asia/Kuwait",
    "Asia/Macao",
    "Asia/Macau",
    "Asia/Magadan",
    "Asia/Makassar",
    "Asia/Manila",
    "Asia/Muscat",
    "Asia/Nicosia",
    "Asia/Novokuznetsk",
    "Asia/Novosibirsk",
    "Asia/Omsk",
    "Asia/Oral",
    "Asia/Phnom_Penh",
    "Asia/Pontianak",
    "Asia/Pyongyang",
    "Asia/Qatar",
    "Asia/Qostanay",
    "Asia/Qyzylorda",
    "Asia/Rangoon",
    "Asia/Riyadh",
    "Asia/Saigon",
    "Asia/Sakhalin",
    "Asia/Samarkand",
    "Asia/Seoul",
    "Asia/Shanghai",
    "Asia/Singapore",
    "Asia/Srednekolymsk",
    "Asia/Taipei",
    "Asia/Tashkent",
    "Asia/Tbilisi",
    "Asia/Tehran",
    "Asia/Tel_Aviv",
    "Asia/Thimbu",
    "Asia/Thimphu",
    "Asia/Tokyo",
    "Asia/Tomsk",
    "Asia/Ujung_Pandang",
    "Asia/Ulaanbaatar",
    "Asia/Ulan_Bator",
    "Asia/Urumqi",
    "Asia/Ust-Nera",
    "Asia/Vientiane",
    "Asia/Vladivostok",
    "Asia/Yakutsk",
    "Asia/Yangon",
    "Asia/Yekaterinburg",
    "Asia/Yerevan",
    "Atlantic/Azores",
    "Atlantic/Bermuda",
    "Atlantic/Canary",
    "Atlantic/Cape_Verde",
    "Atlantic/Faeroe",
    "Atlantic/Faroe",
    "Atlantic/Jan_Mayen",
    "Atlantic/Madeira",
    "Atlantic/Reykjavik",
    "Atlantic/South_Georgia",
    "Atlantic/St_Helena",
    "Atlantic/Stanley",
    "Australia/ACT",
    "Australia/Adelaide",
    "Australia/Brisbane",
    "Australia/Broken_Hill",
    "Australia/Canberra",
    "Australia/Currie",
    "Australia/Darwin",
    "Australia/Eucla",
    "Australia/Hobart",
    "Australia/LHI",
    "Australia/Lindeman",
    "Australia/Lord_Howe",
    "Australia/Melbourne",
    "Australia/NSW",
    "Australia/North",
    "Australia/Perth",
    "Australia/Queensland",
    "Australia/South",
    "Australia/Sydney",
    "Australia/Tasmania",
    "Australia/Victoria",
    "Australia/West",
    "Australia/Yancowinna",
    "Brazil/Acre",
    "Brazil/DeNoronha",
    "Brazil/East",
    "Brazil/West",
    "CET",
    "CST6CDT",
    "Canada/Atlantic",
    "Canada/Central",
    "Canada/Eastern",
    "Canada/Mountain",
    "Canada/Newfoundland",
    "Canada/Pacific",
    "Canada/Saskatchewan",
    "Canada/Yukon",
    "Chile/Continental",
    "Chile/EasterIsland",
    "Cuba",
    "EET",
    "EST",
    "EST5EDT",
    "Egypt",
    "Eire",
    "Etc/GMT",
    "Etc/GMT+0",
    "Etc/GMT+1",
    "Etc/GMT+10",
    "Etc/GMT+11",
    "Etc/GMT+12",
    "Etc/GMT+2",
    "Etc/GMT+3",
    "Etc/GMT+4",
    "Etc/GMT+5",
    "Etc/GMT+6",
    "Etc/GMT+7",
    "Etc/GMT+8",
    "Etc/GMT+9",
    "Etc/GMT-0",
    "Etc/GMT-1",
    "Etc/GMT-10",
    "Etc/GMT-11",
    "Etc/GMT-12",
    "Etc/GMT-13",
    "Etc/GMT-14",
    "Etc/GMT-2",
    "Etc/GMT-3",
    "Etc/GMT-4",
    "Etc/GMT-5",
    "Etc/GMT-6",
    "Etc/GMT-7",
    "Etc/GMT-8",
    "Etc/GMT-9",
    "Etc/GMT0",
    "Etc/Greenwich",
    "Etc/UCT",
    "Etc/UTC",
    "Etc/Universal",
    "Etc/Zulu",
    "Europe/Amsterdam",
    "Europe/Andorra",
    "Europe/Astrakhan",
    "Europe/Athens",
    "Europe/Belfast",
    "Europe/Belgrade",
    "Europe/Berlin",
    "Europe/Bratislava",
    "Europe/Brussels",
    "Europe/Bucharest",
    "Europe/Budapest",
    "Europe/Busingen",
    "Europe/Chisinau",
    "Europe/Copenhagen",
    "Europe/Dublin",
    "Europe/Gibraltar",
    "Europe/Guernsey",
    "Europe/Helsinki",
    "Europe/Isle_of_Man",
    "Europe/Istanbul",
    "Europe/Jersey",
    "Europe/Kaliningrad",
    "Europe/Kiev",
    "Europe/Kirov",
    "Europe/Lisbon",
    "Europe/Ljubljana",
    "Europe/London",
    "Europe/Luxembourg",
    "Europe/Madrid",
    "Europe/Malta",
    "Europe/Mariehamn",
    "Europe/Minsk",
    "Europe/Monaco",
    "Europe/Moscow",
    "Europe/Nicosia",
    "Europe/Oslo",
    "Europe/Paris",
    "Europe/Podgorica",
    "Europe/Prague",
    "Europe/Riga",
    "Europe/Rome",
    "Europe/Samara",
    "Europe/San_Marino",
    "Europe/Sarajevo",
    "Europe/Saratov",
    "Europe/Simferopol",
    "Europe/Skopje",
    "Europe/Sofia",
    "Europe/Stockholm",
    "Europe/Tallinn",
    "Europe/Tirane",
    "Europe/Tiraspol",
    "Europe/Ulyanovsk",
    "Europe/Uzhgorod",
    "Europe/Vaduz",
    "Europe/Vatican",
    "Europe/Vienna",
    "Europe/Vilnius",
    "Europe/Volgograd",
    "Europe/Warsaw",
    "Europe/Zagreb",
    "Europe/Zaporozhye",
    "Europe/Zurich",
    "GB",
    "GB-Eire",
    "GMT",
    "GMT+0",
    "GMT-0",
    "GMT0",
    "Greenwich",
    "HST",
    "Hongkong",
    "Iceland",
    "Indian/Antananarivo",
    "Indian/Chagos",
    "Indian/Christmas",
    "Indian/Cocos",
    "Indian/Comoro",
    "Indian/Kerguelen",
    "Indian/Mahe",
    "Indian/Maldives",
    "Indian/Mauritius",
    "Indian/Mayotte",
    "Indian/Reunion",
    "Iran",
    "Israel",
    "Jamaica",
    "Japan",
    "Kwajalein",
    "Libya",
    "MET",
    "MST",
    "MST7MDT",
    "Mexico/BajaNorte",
    "Mexico/BajaSur",
    "Mexico/General",
    "NZ",
    "NZ-CHAT",
    "Navajo",
    "PRC",
    "PST8PDT",
    "Pacific/Apia",
    "Pacific/Auckland",
    "Pacific/Bougainville",
    "Pacific/Chatham",
    "Pacific/Chuuk",
    "Pacific/Easter",
    "Pacific/Efate",
    "Pacific/Enderbury",
    "Pacific/Fakaofo",
    "Pacific/Fiji",
    "Pacific/Funafuti",
    "Pacific/Galapagos",
    "Pacific/Gambier",
    "Pacific/Guadalcanal",
    "Pacific/Guam",
    "Pacific/Honolulu",
    "Pacific/Johnston",
    "Pacific/Kiritimati",
    "Pacific/Kosrae",
    "Pacific/Kwajalein",
    "Pacific/Majuro",
    "Pacific/Marquesas",
    "Pacific/Midway",
    "Pacific/Nauru",
    "Pacific/Niue",
    "Pacific/Norfolk",
    "Pacific/Noumea",
    "Pacific/Pago_Pago",
    "Pacific/Palau",
    "Pacific/Pitcairn",
    "Pacific/Pohnpei",
    "Pacific/Ponape",
    "Pacific/Port_Moresby",
    "Pacific/Rarotonga",
    "Pacific/Saipan",
    "Pacific/Samoa",
    "Pacific/Tahiti",
    "Pacific/Tarawa",
    "Pacific/Tongatapu",
    "Pacific/Truk",
    "Pacific/Wake",
    "Pacific/Wallis",
    "Pacific/Yap",
    "Poland",
    "Portugal",
    "ROC",
    "ROK",
    "Singapore",
    "Turkey",
    "UCT",
    "US/Alaska",
    "US/Aleutian",
    "US/Arizona",
    "US/Central",
    "US/East-Indiana",
    "US/Eastern",
    "US/Hawaii",
    "US/Indiana-Starke",
    "US/Michigan",
    "US/Mountain",
    "US/Pacific",
    "US/Samoa",
    "UTC",
    "Universal",
    "W-SU",
    "WET",
    "Zulu",
)



"""
    sphinx.util.jsonimpl
    ~~~~~~~~~~~~~~~~~~~~

    JSON serializer implementation wrapper.

    :copyright: Copyright 2007-2019 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import json
from collections import UserString

if False:
    # For type annotation
    from typing import Any, IO  # NOQA


class SphinxJSONEncoder(json.JSONEncoder):
    """JSONEncoder subclass that forces translation proxies."""
    def default(self, obj):
        # type: (Any) -> str
        if isinstance(obj, UserString):
            return str(obj)
        return super().default(obj)


def dump(obj, fp, *args, **kwds):
    # type: (Any, IO, Any, Any) -> None
    kwds['cls'] = SphinxJSONEncoder
    json.dump(obj, fp, *args, **kwds)


def dumps(obj, *args, **kwds):
    # type: (Any, Any, Any) -> str
    kwds['cls'] = SphinxJSONEncoder
    return json.dumps(obj, *args, **kwds)


def load(*args, **kwds):
    # type: (Any, Any) -> Any
    return json.load(*args, **kwds)


def loads(*args, **kwds):
    # type: (Any, Any) -> Any
    return json.loads(*args, **kwds)



# Generated by Snowball 2.2.0 - https://snowballstem.org/

from .basestemmer import BaseStemmer
from .among import Among


class TamilStemmer(BaseStemmer):
    '''
    This class implements the stemming algorithm defined by a snowball script.
    Generated by Snowball 2.2.0 - https://snowballstem.org/
    '''

    a_0 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B99", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9E", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BA8", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BAE", -1, -1),
        Among(u"\u0BAF", -1, -1),
        Among(u"\u0BB5", -1, -1)
    ]

    a_1 = [
        Among(u"\u0BA8\u0BCD\u0BA4", -1, -1),
        Among(u"\u0BA8\u0BCD\u0BA4\u0BCD", -1, -1),
        Among(u"\u0BA8\u0BCD", -1, -1)
    ]

    a_2 = [
        Among(u"\u0BBF", -1, -1),
        Among(u"\u0BC0", -1, -1),
        Among(u"\u0BC8", -1, -1)
    ]

    a_3 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9F", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BB1", -1, -1)
    ]

    a_4 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9F", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BB1", -1, -1)
    ]

    a_5 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9F", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BB1", -1, -1)
    ]

    a_6 = [
        Among(u"\u0BAF", -1, -1),
        Among(u"\u0BB0", -1, -1),
        Among(u"\u0BB2", -1, -1),
        Among(u"\u0BB3", -1, -1),
        Among(u"\u0BB4", -1, -1),
        Among(u"\u0BB5", -1, -1)
    ]

    a_7 = [
        Among(u"\u0B99", -1, -1),
        Among(u"\u0B9E", -1, -1),
        Among(u"\u0BA3", -1, -1),
        Among(u"\u0BA8", -1, -1),
        Among(u"\u0BA9", -1, -1),
        Among(u"\u0BAE", -1, -1)
    ]

    a_8 = [
        Among(u"\u0BAF", -1, -1),
        Among(u"\u0BB5", -1, -1),
        Among(u"\u0BB5\u0BCD", -1, -1)
    ]

    a_9 = [
        Among(u"\u0BBE", -1, -1),
        Among(u"\u0BBF", -1, -1),
        Among(u"\u0BC0", -1, -1),
        Among(u"\u0BC1", -1, -1),
        Among(u"\u0BC2", -1, -1),
        Among(u"\u0BC6", -1, -1),
        Among(u"\u0BC7", -1, -1),
        Among(u"\u0BC8", -1, -1)
    ]

    a_10 = [
        Among(u"\u0BBE", -1, -1),
        Among(u"\u0BBF", -1, -1),
        Among(u"\u0BC0", -1, -1),
        Among(u"\u0BC1", -1, -1),
        Among(u"\u0BC2", -1, -1),
        Among(u"\u0BC6", -1, -1),
        Among(u"\u0BC7", -1, -1),
        Among(u"\u0BC8", -1, -1)
    ]

    a_11 = [
        Among(u"\u0B85", -1, -1),
        Among(u"\u0B87", -1, -1),
        Among(u"\u0B89", -1, -1)
    ]

    a_12 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B99", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9E", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BA8", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BAE", -1, -1),
        Among(u"\u0BAF", -1, -1),
        Among(u"\u0BB5", -1, -1)
    ]

    a_13 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9F", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BB1", -1, -1)
    ]

    a_14 = [
        Among(u"\u0BBE", -1, -1),
        Among(u"\u0BC7", -1, -1),
        Among(u"\u0BCB", -1, -1)
    ]

    a_15 = [
        Among(u"\u0BAA\u0BBF", -1, -1),
        Among(u"\u0BB5\u0BBF", -1, -1)
    ]

    a_16 = [
        Among(u"\u0BBE", -1, -1),
        Among(u"\u0BBF", -1, -1),
        Among(u"\u0BC0", -1, -1),
        Among(u"\u0BC1", -1, -1),
        Among(u"\u0BC2", -1, -1),
        Among(u"\u0BC6", -1, -1),
        Among(u"\u0BC7", -1, -1),
        Among(u"\u0BC8", -1, -1)
    ]

    a_17 = [
        Among(u"\u0BAA\u0B9F\u0BCD\u0B9F", -1, -1),
        Among(u"\u0BAA\u0B9F\u0BCD\u0B9F\u0BA3", -1, -1),
        Among(u"\u0BA4\u0BBE\u0BA9", -1, -1),
        Among(u"\u0BAA\u0B9F\u0BBF\u0BA4\u0BBE\u0BA9", 2, -1),
        Among(u"\u0B95\u0BC1\u0BB0\u0BBF\u0BAF", -1, -1),
        Among(u"\u0BAA\u0B9F\u0BBF", -1, -1),
        Among(u"\u0BAA\u0BB1\u0BCD\u0BB1\u0BBF", -1, -1),
        Among(u"\u0BAA\u0B9F\u0BC1", -1, -1),
        Among(u"\u0BB5\u0BBF\u0B9F\u0BC1", -1, -1),
        Among(u"\u0BAA\u0B9F\u0BCD\u0B9F\u0BC1", -1, -1),
        Among(u"\u0BB5\u0BBF\u0B9F\u0BCD\u0B9F\u0BC1", -1, -1),
        Among(u"\u0BAA\u0B9F\u0BCD\u0B9F\u0BA4\u0BC1", -1, -1),
        Among(u"\u0BC6\u0BB2\u0BCD\u0BB2\u0BBE\u0BAE\u0BCD", -1, -1)
    ]

    a_18 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9F", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BB1", -1, -1)
    ]

    a_19 = [
        Among(u"\u0B95", -1, -1),
        Among(u"\u0B9A", -1, -1),
        Among(u"\u0B9F", -1, -1),
        Among(u"\u0BA4", -1, -1),
        Among(u"\u0BAA", -1, -1),
        Among(u"\u0BB1", -1, -1)
    ]

    a_20 = [
        Among(u"\u0BBE", -1, -1),
        Among(u"\u0BBF", -1, -1),
        Among(u"\u0BC0", -1, -1),
        Among(u"\u0BC1", -1, -1),
        Among(u"\u0BC2", -1, -1),
        Among(u"\u0BC6", -1, -1),
        Among(u"\u0BC7", -1, -1),
        Among(u"\u0BC8", -1, -1)
    ]

    a_21 = [
        Among(u"\u0BBE", -1, -1),
        Among(u"\u0BBF", -1, -1),
        Among(u"\u0BC0", -1, -1),
        Among(u"\u0BC1", -1, -1),
        Among(u"\u0BC2", -1, -1),
        Among(u"\u0BC6", -1, -1),
        Among(u"\u0BC7", -1, -1),
        Among(u"\u0BC8", -1, -1)
    ]

    a_22 = [
        Among(u"\u0BAA\u0B9F\u0BC1", -1, -1),
        Among(u"\u0B95\u0BCA\u0BA3\u0BCD\u0B9F\u0BBF\u0BB0\u0BCD", -1, -1)
    ]

    a_23 = [
        Among(u"\u0B85", -1, -1),
        Among(u"\u0B86", -1, -1),
        Among(u"\u0B87", -1, -1),
        Among(u"\u0B88", -1, -1),
        Among(u"\u0B89", -1, -1),
        Among(u"\u0B8A", -1, -1),
        Among(u"\u0B8E", -1, -1),
        Among(u"\u0B8F", -1, -1),
        Among(u"\u0B90", -1, -1),
        Among(u"\u0B92", -1, -1),
        Among(u"\u0B93", -1, -1),
        Among(u"\u0B94", -1, -1)
    ]

    a_24 = [
        Among(u"\u0BBE", -1, -1),
        Among(u"\u0BBF", -1, -1),
        Among(u"\u0BC0", -1, -1),
        Among(u"\u0BC1", -1, -1),
        Among(u"\u0BC2", -1, -1),
        Among(u"\u0BC6", -1, -1),
        Among(u"\u0BC7", -1, -1),
        Among(u"\u0BC8", -1, -1)
    ]

    a_25 = [
        Among(u"\u0B95\u0BBF\u0BB1", -1, -1),
        Among(u"\u0B95\u0BBF\u0BA9\u0BCD\u0BB1", -1, -1),
        Among(u"\u0BBE\u0BA8\u0BBF\u0BA9\u0BCD\u0BB1", -1, -1),
        Among(u"\u0B95\u0BBF\u0BB1\u0BCD", -1, -1),
        Among(u"\u0B95\u0BBF\u0BA9\u0BCD\u0BB1\u0BCD", -1, -1),
        Among(u"\u0BBE\u0BA8\u0BBF\u0BA9\u0BCD\u0BB1\u0BCD", -1, -1)
    ]

    B_found_vetrumai_urupu = False
    B_found_a_match = False

    def __r_has_min_length(self):
        if not len(self.current) > 4:
            return False
        return True

    def __r_fix_va_start(self):
        try:
            v_1 = self.cursor
            try:
                v_2 = self.cursor
                v_3 = self.cursor
                try:
                    if not self.eq_s(u"\u0BB5\u0BCB"):
                        self.cursor = v_3
                        raise lab2()
                except lab2: pass
                self.cursor = v_2
                self.bra = self.cursor
                if not self.eq_s(u"\u0BB5\u0BCB"):
                    raise lab1()
                self.ket = self.cursor
                if not self.slice_from(u"\u0B93"):
                    return False
                raise lab0()
            except lab1: pass
            self.cursor = v_1
            try:
                v_4 = self.cursor
                v_5 = self.cursor
                try:
                    if not self.eq_s(u"\u0BB5\u0BCA"):
                        self.cursor = v_5
                        raise lab4()
                except lab4: pass
                self.cursor = v_4
                self.bra = self.cursor
                if not self.eq_s(u"\u0BB5\u0BCA"):
                    raise lab3()
                self.ket = self.cursor
                if not self.slice_from(u"\u0B92"):
                    return False
                raise lab0()
            except lab3: pass
            self.cursor = v_1
            try:
                v_6 = self.cursor
                v_7 = self.cursor
                try:
                    if not self.eq_s(u"\u0BB5\u0BC1"):
                        self.cursor = v_7
                        raise lab6()
                except lab6: pass
                self.cursor = v_6
                self.bra = self.cursor
                if not self.eq_s(u"\u0BB5\u0BC1"):
                    raise lab5()
                self.ket = self.cursor
                if not self.slice_from(u"\u0B89"):
                    return False
                raise lab0()
            except lab5: pass
            self.cursor = v_1
            v_8 = self.cursor
            v_9 = self.cursor
            try:
                if not self.eq_s(u"\u0BB5\u0BC2"):
                    self.cursor = v_9
                    raise lab7()
            except lab7: pass
            self.cursor = v_8
            self.bra = self.cursor
            if not self.eq_s(u"\u0BB5\u0BC2"):
                return False
            self.ket = self.cursor
            if not self.slice_from(u"\u0B8A"):
                return False
        except lab0: pass
        return True

    def __r_fix_endings(self):
        v_1 = self.cursor
        try:
            while True:
                v_2 = self.cursor
                try:
                    if not self.__r_fix_ending():
                        raise lab1()
                    continue
                except lab1: pass
                self.cursor = v_2
                break
        except lab0: pass
        self.cursor = v_1
        return True

    def __r_remove_question_prefixes(self):
        self.bra = self.cursor
        if not self.eq_s(u"\u0B8E"):
            return False
        if self.find_among(TamilStemmer.a_0) == 0:
            return False
        if not self.eq_s(u"\u0BCD"):
            return False
        self.ket = self.cursor
        if not self.slice_del():
            return False

        v_1 = self.cursor
        self.__r_fix_va_start()
        self.cursor = v_1
        return True

    def __r_fix_ending(self):
        if not len(self.current) > 3:
            return False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        try:
            v_1 = self.limit - self.cursor
            try:
                self.ket = self.cursor
                if self.find_among_b(TamilStemmer.a_1) == 0:
                    raise lab1()
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                raise lab0()
            except lab1: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BAF\u0BCD"):
                    raise lab2()
                v_2 = self.limit - self.cursor
                if self.find_among_b(TamilStemmer.a_2) == 0:
                    raise lab2()
                self.cursor = self.limit - v_2
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                raise lab0()
            except lab2: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                try:
                    v_3 = self.limit - self.cursor
                    try:
                        if not self.eq_s_b(u"\u0B9F\u0BCD\u0BAA\u0BCD"):
                            raise lab5()
                        raise lab4()
                    except lab5: pass
                    self.cursor = self.limit - v_3
                    if not self.eq_s_b(u"\u0B9F\u0BCD\u0B95\u0BCD"):
                        raise lab3()
                except lab4: pass
                self.bra = self.cursor
                if not self.slice_from(u"\u0BB3\u0BCD"):
                    return False
                raise lab0()
            except lab3: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BA9\u0BCD\u0BB1\u0BCD"):
                    raise lab6()
                self.bra = self.cursor
                if not self.slice_from(u"\u0BB2\u0BCD"):
                    return False
                raise lab0()
            except lab6: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BB1\u0BCD\u0B95\u0BCD"):
                    raise lab7()
                self.bra = self.cursor
                if not self.slice_from(u"\u0BB2\u0BCD"):
                    return False
                raise lab0()
            except lab7: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0B9F\u0BCD\u0B9F\u0BCD"):
                    raise lab8()
                self.bra = self.cursor
                if not self.slice_from(u"\u0B9F\u0BC1"):
                    return False
                raise lab0()
            except lab8: pass
            self.cursor = self.limit - v_1
            try:
                if not self.B_found_vetrumai_urupu:
                    raise lab9()
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BA4\u0BCD\u0BA4\u0BCD"):
                    raise lab9()
                v_4 = self.limit - self.cursor
                v_5 = self.limit - self.cursor
                try:
                    if not self.eq_s_b(u"\u0BC8"):
                        raise lab10()
                    raise lab9()
                except lab10: pass
                self.cursor = self.limit - v_5
                self.cursor = self.limit - v_4
                self.bra = self.cursor
                if not self.slice_from(u"\u0BAE\u0BCD"):
                    return False
                self.bra = self.cursor
                raise lab0()
            except lab9: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                try:
                    v_6 = self.limit - self.cursor
                    try:
                        if not self.eq_s_b(u"\u0BC1\u0B95\u0BCD"):
                            raise lab13()
                        raise lab12()
                    except lab13: pass
                    self.cursor = self.limit - v_6
                    if not self.eq_s_b(u"\u0BC1\u0B95\u0BCD\u0B95\u0BCD"):
                        raise lab11()
                except lab12: pass
                self.bra = self.cursor
                if not self.slice_from(u"\u0BCD"):
                    return False
                raise lab0()
            except lab11: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BCD"):
                    raise lab14()
                if self.find_among_b(TamilStemmer.a_3) == 0:
                    raise lab14()
                if not self.eq_s_b(u"\u0BCD"):
                    raise lab14()
                if self.find_among_b(TamilStemmer.a_4) == 0:
                    raise lab14()
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                raise lab0()
            except lab14: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BC1\u0B95\u0BCD"):
                    raise lab15()
                self.bra = self.cursor
                if not self.slice_from(u"\u0BCD"):
                    return False
                raise lab0()
            except lab15: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BCD"):
                    raise lab16()
                if self.find_among_b(TamilStemmer.a_5) == 0:
                    raise lab16()
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                raise lab0()
            except lab16: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BCD"):
                    raise lab17()
                try:
                    v_7 = self.limit - self.cursor
                    try:
                        if self.find_among_b(TamilStemmer.a_6) == 0:
                            raise lab19()
                        raise lab18()
                    except lab19: pass
                    self.cursor = self.limit - v_7
                    if self.find_among_b(TamilStemmer.a_7) == 0:
                        raise lab17()
                except lab18: pass
                if not self.eq_s_b(u"\u0BCD"):
                    raise lab17()
                self.bra = self.cursor
                if not self.slice_from(u"\u0BCD"):
                    return False
                raise lab0()
            except lab17: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if self.find_among_b(TamilStemmer.a_8) == 0:
                    raise lab20()
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                raise lab0()
            except lab20: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BA9\u0BC1"):
                    raise lab21()
                v_8 = self.limit - self.cursor
                v_9 = self.limit - self.cursor
                try:
                    if self.find_among_b(TamilStemmer.a_9) == 0:
                        raise lab22()
                    raise lab21()
                except lab22: pass
                self.cursor = self.limit - v_9
                self.cursor = self.limit - v_8
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                raise lab0()
            except lab21: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0B99\u0BCD"):
                    raise lab23()
                v_10 = self.limit - self.cursor
                v_11 = self.limit - self.cursor
                try:
                    if not self.eq_s_b(u"\u0BC8"):
                        raise lab24()
                    raise lab23()
                except lab24: pass
                self.cursor = self.limit - v_11
                self.cursor = self.limit - v_10
                self.bra = self.cursor
                if not self.slice_from(u"\u0BAE\u0BCD"):
                    return False
                raise lab0()
            except lab23: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0B99\u0BCD"):
                    raise lab25()
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                raise lab0()
            except lab25: pass
            self.cursor = self.limit - v_1
            self.ket = self.cursor
            if not self.eq_s_b(u"\u0BCD"):
                return False
            v_12 = self.limit - self.cursor
            try:
                v_13 = self.limit - self.cursor
                try:
                    if self.find_among_b(TamilStemmer.a_10) == 0:
                        raise lab27()
                    raise lab26()
                except lab27: pass
                self.cursor = self.limit - v_13
                if not self.eq_s_b(u"\u0BCD"):
                    return False
            except lab26: pass
            self.cursor = self.limit - v_12
            self.bra = self.cursor
            if not self.slice_del():
                return False

        except lab0: pass
        self.cursor = self.limit_backward
        return True

    def __r_remove_pronoun_prefixes(self):
        self.B_found_a_match = False
        self.bra = self.cursor
        if self.find_among(TamilStemmer.a_11) == 0:
            return False
        if self.find_among(TamilStemmer.a_12) == 0:
            return False
        if not self.eq_s(u"\u0BCD"):
            return False
        self.ket = self.cursor
        if not self.slice_del():
            return False

        self.B_found_a_match = True
        v_1 = self.cursor
        self.__r_fix_va_start()
        self.cursor = v_1
        return True

    def __r_remove_plural_suffix(self):
        self.B_found_a_match = False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        try:
            v_1 = self.limit - self.cursor
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BC1\u0B99\u0BCD\u0B95\u0BB3\u0BCD"):
                    raise lab1()
                v_2 = self.limit - self.cursor
                v_3 = self.limit - self.cursor
                try:
                    if self.find_among_b(TamilStemmer.a_13) == 0:
                        raise lab2()
                    raise lab1()
                except lab2: pass
                self.cursor = self.limit - v_3
                self.cursor = self.limit - v_2
                self.bra = self.cursor
                if not self.slice_from(u"\u0BCD"):
                    return False
                raise lab0()
            except lab1: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BB1\u0BCD\u0B95\u0BB3\u0BCD"):
                    raise lab3()
                self.bra = self.cursor
                if not self.slice_from(u"\u0BB2\u0BCD"):
                    return False
                raise lab0()
            except lab3: pass
            self.cursor = self.limit - v_1
            try:
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0B9F\u0BCD\u0B95\u0BB3\u0BCD"):
                    raise lab4()
                self.bra = self.cursor
                if not self.slice_from(u"\u0BB3\u0BCD"):
                    return False
                raise lab0()
            except lab4: pass
            self.cursor = self.limit - v_1
            self.ket = self.cursor
            if not self.eq_s_b(u"\u0B95\u0BB3\u0BCD"):
                return False
            self.bra = self.cursor
            if not self.slice_del():
                return False

        except lab0: pass
        self.B_found_a_match = True
        self.cursor = self.limit_backward
        return True

    def __r_remove_question_suffixes(self):
        if not self.__r_has_min_length():
            return False
        self.B_found_a_match = False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        v_1 = self.limit - self.cursor
        try:
            self.ket = self.cursor
            if self.find_among_b(TamilStemmer.a_14) == 0:
                raise lab0()
            self.bra = self.cursor
            if not self.slice_from(u"\u0BCD"):
                return False
            self.B_found_a_match = True
        except lab0: pass
        self.cursor = self.limit - v_1
        self.cursor = self.limit_backward
        self.__r_fix_endings()
        return True

    def __r_remove_command_suffixes(self):
        if not self.__r_has_min_length():
            return False
        self.B_found_a_match = False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        self.ket = self.cursor
        if self.find_among_b(TamilStemmer.a_15) == 0:
            return False
        self.bra = self.cursor
        if not self.slice_del():
            return False

        self.B_found_a_match = True
        self.cursor = self.limit_backward
        return True

    def __r_remove_um(self):
        self.B_found_a_match = False
        if not self.__r_has_min_length():
            return False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        self.ket = self.cursor
        if not self.eq_s_b(u"\u0BC1\u0BAE\u0BCD"):
            return False
        self.bra = self.cursor
        if not self.slice_from(u"\u0BCD"):
            return False
        self.B_found_a_match = True
        self.cursor = self.limit_backward
        v_1 = self.cursor
        self.__r_fix_ending()
        self.cursor = v_1
        return True

    def __r_remove_common_word_endings(self):
        self.B_found_a_match = False
        if not self.__r_has_min_length():
            return False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        try:
            v_1 = self.limit - self.cursor
            try:
                v_2 = self.limit - self.cursor
                self.ket = self.cursor
                try:
                    v_3 = self.limit - self.cursor
                    try:
                        if not self.eq_s_b(u"\u0BC1\u0B9F\u0BA9\u0BCD"):
                            raise lab3()
                        raise lab2()
                    except lab3: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0BB2\u0BCD\u0BB2\u0BC8"):
                            raise lab4()
                        raise lab2()
                    except lab4: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0B9F\u0BAE\u0BCD"):
                            raise lab5()
                        raise lab2()
                    except lab5: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0BA9\u0BCD\u0BB1\u0BBF"):
                            raise lab6()
                        raise lab2()
                    except lab6: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BBE\u0B95\u0BBF"):
                            raise lab7()
                        raise lab2()
                    except lab7: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BBE\u0B95\u0BBF\u0BAF"):
                            raise lab8()
                        raise lab2()
                    except lab8: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BC6\u0BA9\u0BCD\u0BB1\u0BC1"):
                            raise lab9()
                        raise lab2()
                    except lab9: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BC1\u0BB3\u0BCD\u0BB3"):
                            raise lab10()
                        raise lab2()
                    except lab10: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BC1\u0B9F\u0BC8\u0BAF"):
                            raise lab11()
                        raise lab2()
                    except lab11: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BC1\u0B9F\u0BC8"):
                            raise lab12()
                        raise lab2()
                    except lab12: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BC6\u0BA9\u0BC1\u0BAE\u0BCD"):
                            raise lab13()
                        raise lab2()
                    except lab13: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BB2\u0BCD\u0BB2"):
                            raise lab14()
                        v_4 = self.limit - self.cursor
                        v_5 = self.limit - self.cursor
                        try:
                            if self.find_among_b(TamilStemmer.a_16) == 0:
                                raise lab15()
                            raise lab14()
                        except lab15: pass
                        self.cursor = self.limit - v_5
                        self.cursor = self.limit - v_4
                        raise lab2()
                    except lab14: pass
                    self.cursor = self.limit - v_3
                    try:
                        if not self.eq_s_b(u"\u0BC6\u0BA9"):
                            raise lab16()
                        raise lab2()
                    except lab16: pass
                    self.cursor = self.limit - v_3
                    if not self.eq_s_b(u"\u0BBE\u0B95\u0BBF"):
                        raise lab1()
                except lab2: pass
                self.bra = self.cursor
                if not self.slice_from(u"\u0BCD"):
                    return False
                self.B_found_a_match = True
                self.cursor = self.limit - v_2
                raise lab0()
            except lab1: pass
            self.cursor = self.limit - v_1
            v_6 = self.limit - self.cursor
            self.ket = self.cursor
            if self.find_among_b(TamilStemmer.a_17) == 0:
                return False
            self.bra = self.cursor
            if not self.slice_del():
                return False

            self.B_found_a_match = True
            self.cursor = self.limit - v_6
        except lab0: pass
        self.cursor = self.limit_backward
        self.__r_fix_endings()
        return True

    def __r_remove_vetrumai_urupukal(self):
        self.B_found_a_match = False
        self.B_found_vetrumai_urupu = False
        if not self.__r_has_min_length():
            return False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        try:
            v_1 = self.limit - self.cursor
            try:
                v_2 = self.limit - self.cursor
                self.ket = self.cursor
                if not self.eq_s_b(u"\u0BA9\u0BC8"):
                    raise lab1()
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                self.cursor = self.limit - v_2
                raise lab0()
            except lab1: pass
            self.cursor = self.limit - v_1
            try:
                v_3 = self.limit - self.cursor
                self.ket = self.cursor
                try:
                    v_4 = self.limit - self.cursor
                    try:
                        try:
                            v_5 = self.limit - self.cursor
                            try:
                                if not self.eq_s_b(u"\u0BBF\u0BA9\u0BC8"):
                                    raise lab6()
                                raise lab5()
                            except lab6: pass
                            self.cursor = self.limit - v_5
                            if not self.eq_s_b(u"\u0BC8"):
                                raise lab4()
                        except lab5: pass
                        v_6 = self.limit - self.cursor
                        v_7 = self.limit - self.cursor
                        try:
                            if self.find_among_b(TamilStemmer.a_18) == 0:
                                raise lab7()
                            raise lab4()
                        except lab7: pass
                        self.cursor = self.limit - v_7
                        self.cursor = self.limit - v_6
                        raise lab3()
                    except lab4: pass
                    self.cursor = self.limit - v_4
                    if not self.eq_s_b(u"\u0BC8"):
                        raise lab2()
                    v_8 = self.limit - self.cursor
                    if self.find_among_b(TamilStemmer.a_19) == 0:
                        raise lab2()
                    if not self.eq_s_b(u"\u0BCD"):
                        raise lab2()
                    self.cursor = self.limit - v_8
                except lab3: pass
                self.bra = self.cursor
                if not self.slice_from(u"\u0BCD"):
                    return False
                self.cursor = self.limit - v_3
                raise lab0()
            except lab2: pass
            self.cursor = self.limit - v_1
            try:
                v_9 = self.limit - self.cursor
                self.ket = self.cursor
                try:
                    v_10 = self.limit - self.cursor
                    try:
                        if not self.eq_s_b(u"\u0BCA\u0B9F\u0BC1"):
                            raise lab10()
                        raise lab9()
                    except lab10: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BCB\u0B9F\u0BC1"):
                            raise lab11()
                        raise lab9()
                    except lab11: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0BB2\u0BCD"):
                            raise lab12()
                        raise lab9()
                    except lab12: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0BB1\u0BCD"):
                            raise lab13()
                        raise lab9()
                    except lab13: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0BA9\u0BCD"):
                            raise lab14()
                        v_11 = self.limit - self.cursor
                        v_12 = self.limit - self.cursor
                        try:
                            if not self.eq_s_b(u"\u0BAE"):
                                raise lab15()
                            raise lab14()
                        except lab15: pass
                        self.cursor = self.limit - v_12
                        self.cursor = self.limit - v_11
                        raise lab9()
                    except lab14: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0BA9\u0BCD\u0BB1\u0BC1"):
                            raise lab16()
                        raise lab9()
                    except lab16: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BBF\u0BB0\u0BC1\u0BA8\u0BCD\u0BA4\u0BC1"):
                            raise lab17()
                        raise lab9()
                    except lab17: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BB5\u0BBF\u0B9F"):
                            raise lab18()
                        raise lab9()
                    except lab18: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not len(self.current) >= 7:
                            raise lab19()
                        if not self.eq_s_b(u"\u0BBF\u0B9F\u0BAE\u0BCD"):
                            raise lab19()
                        raise lab9()
                    except lab19: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BBE\u0BB2\u0BCD"):
                            raise lab20()
                        raise lab9()
                    except lab20: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BC1\u0B9F\u0BC8"):
                            raise lab21()
                        raise lab9()
                    except lab21: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BBE\u0BAE\u0BB2\u0BCD"):
                            raise lab22()
                        raise lab9()
                    except lab22: pass
                    self.cursor = self.limit - v_10
                    try:
                        if not self.eq_s_b(u"\u0BB2\u0BCD"):
                            raise lab23()
                        v_13 = self.limit - self.cursor
                        v_14 = self.limit - self.cursor
                        try:
                            if self.find_among_b(TamilStemmer.a_20) == 0:
                                raise lab24()
                            raise lab23()
                        except lab24: pass
                        self.cursor = self.limit - v_14
                        self.cursor = self.limit - v_13
                        raise lab9()
                    except lab23: pass
                    self.cursor = self.limit - v_10
                    if not self.eq_s_b(u"\u0BC1\u0BB3\u0BCD"):
                        raise lab8()
                except lab9: pass
                self.bra = self.cursor
                if not self.slice_from(u"\u0BCD"):
                    return False
                self.cursor = self.limit - v_9
                raise lab0()
            except lab8: pass
            self.cursor = self.limit - v_1
            try:
                v_15 = self.limit - self.cursor
                self.ket = self.cursor
                try:
                    v_16 = self.limit - self.cursor
                    try:
                        if not self.eq_s_b(u"\u0B95\u0BA3\u0BCD"):
                            raise lab27()
                        raise lab26()
                    except lab27: pass
                    self.cursor = self.limit - v_16
                    try:
                        if not self.eq_s_b(u"\u0BAE\u0BC1\u0BA9\u0BCD"):
                            raise lab28()
                        raise lab26()
                    except lab28: pass
                    self.cursor = self.limit - v_16
                    try:
                        if not self.eq_s_b(u"\u0BAE\u0BC7\u0BB2\u0BCD"):
                            raise lab29()
                        raise lab26()
                    except lab29: pass
                    self.cursor = self.limit - v_16
                    try:
                        if not self.eq_s_b(u"\u0BAE\u0BC7\u0BB1\u0BCD"):
                            raise lab30()
                        raise lab26()
                    except lab30: pass
                    self.cursor = self.limit - v_16
                    try:
                        if not self.eq_s_b(u"\u0B95\u0BC0\u0BB4\u0BCD"):
                            raise lab31()
                        raise lab26()
                    except lab31: pass
                    self.cursor = self.limit - v_16
                    try:
                        if not self.eq_s_b(u"\u0BAA\u0BBF\u0BA9\u0BCD"):
                            raise lab32()
                        raise lab26()
                    except lab32: pass
                    self.cursor = self.limit - v_16
                    if not self.eq_s_b(u"\u0BA4\u0BC1"):
                        raise lab25()
                    v_17 = self.limit - self.cursor
                    v_18 = self.limit - self.cursor
                    try:
                        if self.find_among_b(TamilStemmer.a_21) == 0:
                            raise lab33()
                        raise lab25()
                    except lab33: pass
                    self.cursor = self.limit - v_18
                    self.cursor = self.limit - v_17
                except lab26: pass
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                self.cursor = self.limit - v_15
                raise lab0()
            except lab25: pass
            self.cursor = self.limit - v_1
            v_19 = self.limit - self.cursor
            self.ket = self.cursor
            if not self.eq_s_b(u"\u0BC0"):
                return False
            self.bra = self.cursor
            if not self.slice_from(u"\u0BBF"):
                return False
            self.cursor = self.limit - v_19
        except lab0: pass
        self.B_found_a_match = True
        self.B_found_vetrumai_urupu = True
        v_20 = self.limit - self.cursor
        try:
            self.ket = self.cursor
            if not self.eq_s_b(u"\u0BBF\u0BA9\u0BCD"):
                raise lab34()
            self.bra = self.cursor
            if not self.slice_from(u"\u0BCD"):
                return False
        except lab34: pass
        self.cursor = self.limit - v_20
        self.cursor = self.limit_backward
        self.__r_fix_endings()
        return True

    def __r_remove_tense_suffixes(self):
        self.B_found_a_match = True
        while True:
            v_1 = self.cursor
            try:
                if not self.B_found_a_match:
                    raise lab0()
                v_2 = self.cursor
                self.__r_remove_tense_suffix()
                self.cursor = v_2
                continue
            except lab0: pass
            self.cursor = v_1
            break
        return True

    def __r_remove_tense_suffix(self):
        self.B_found_a_match = False
        if not self.__r_has_min_length():
            return False
        self.limit_backward = self.cursor
        self.cursor = self.limit
        v_1 = self.limit - self.cursor
        try:
            try:
                v_2 = self.limit - self.cursor
                try:
                    v_3 = self.limit - self.cursor
                    self.ket = self.cursor
                    if self.find_among_b(TamilStemmer.a_22) == 0:
                        raise lab2()
                    self.bra = self.cursor
                    if not self.slice_del():
                        return False

                    self.B_found_a_match = True
                    self.cursor = self.limit - v_3
                    raise lab1()
                except lab2: pass
                self.cursor = self.limit - v_2
                try:
                    v_4 = self.limit - self.cursor
                    self.ket = self.cursor
                    try:
                        v_5 = self.limit - self.cursor
                        try:
                            if not self.eq_s_b(u"\u0BAE\u0BBE\u0BB0\u0BCD"):
                                raise lab5()
                            raise lab4()
                        except lab5: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BAE\u0BBF\u0BA9\u0BCD"):
                                raise lab6()
                            raise lab4()
                        except lab6: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BA9\u0BCD"):
                                raise lab7()
                            raise lab4()
                        except lab7: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BBE\u0BA9\u0BCD"):
                                raise lab8()
                            raise lab4()
                        except lab8: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BBE\u0BB3\u0BCD"):
                                raise lab9()
                            raise lab4()
                        except lab9: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BBE\u0BB0\u0BCD"):
                                raise lab10()
                            raise lab4()
                        except lab10: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BB5\u0BA9\u0BCD"):
                                raise lab11()
                            v_6 = self.limit - self.cursor
                            v_7 = self.limit - self.cursor
                            try:
                                if self.find_among_b(TamilStemmer.a_23) == 0:
                                    raise lab12()
                                raise lab11()
                            except lab12: pass
                            self.cursor = self.limit - v_7
                            self.cursor = self.limit - v_6
                            raise lab4()
                        except lab11: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BB3\u0BCD"):
                                raise lab13()
                            raise lab4()
                        except lab13: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BB5\u0BB3\u0BCD"):
                                raise lab14()
                            raise lab4()
                        except lab14: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BB0\u0BCD"):
                                raise lab15()
                            raise lab4()
                        except lab15: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BB5\u0BB0\u0BCD"):
                                raise lab16()
                            raise lab4()
                        except lab16: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9"):
                                raise lab17()
                            raise lab4()
                        except lab17: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BAA"):
                                raise lab18()
                            raise lab4()
                        except lab18: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0B95"):
                                raise lab19()
                            raise lab4()
                        except lab19: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA4"):
                                raise lab20()
                            raise lab4()
                        except lab20: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BAF"):
                                raise lab21()
                            raise lab4()
                        except lab21: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BAA\u0BA9\u0BCD"):
                                raise lab22()
                            raise lab4()
                        except lab22: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BAA\u0BB3\u0BCD"):
                                raise lab23()
                            raise lab4()
                        except lab23: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BAA\u0BB0\u0BCD"):
                                raise lab24()
                            raise lab4()
                        except lab24: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA4\u0BC1"):
                                raise lab25()
                            v_8 = self.limit - self.cursor
                            v_9 = self.limit - self.cursor
                            try:
                                if self.find_among_b(TamilStemmer.a_24) == 0:
                                    raise lab26()
                                raise lab25()
                            except lab26: pass
                            self.cursor = self.limit - v_9
                            self.cursor = self.limit - v_8
                            raise lab4()
                        except lab25: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BBF\u0BB1\u0BCD\u0BB1\u0BC1"):
                                raise lab27()
                            raise lab4()
                        except lab27: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BAA\u0BAE\u0BCD"):
                                raise lab28()
                            raise lab4()
                        except lab28: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BAE\u0BCD"):
                                raise lab29()
                            raise lab4()
                        except lab29: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA4\u0BC1\u0BAE\u0BCD"):
                                raise lab30()
                            raise lab4()
                        except lab30: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BB1\u0BC1\u0BAE\u0BCD"):
                                raise lab31()
                            raise lab4()
                        except lab31: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0B95\u0BC1\u0BAE\u0BCD"):
                                raise lab32()
                            raise lab4()
                        except lab32: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BC6\u0BA9\u0BCD"):
                                raise lab33()
                            raise lab4()
                        except lab33: pass
                        self.cursor = self.limit - v_5
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BC8"):
                                raise lab34()
                            raise lab4()
                        except lab34: pass
                        self.cursor = self.limit - v_5
                        if not self.eq_s_b(u"\u0BB5\u0BC8"):
                            raise lab3()
                    except lab4: pass
                    self.bra = self.cursor
                    if not self.slice_del():
                        return False

                    self.B_found_a_match = True
                    self.cursor = self.limit - v_4
                    raise lab1()
                except lab3: pass
                self.cursor = self.limit - v_2
                try:
                    v_10 = self.limit - self.cursor
                    self.ket = self.cursor
                    try:
                        v_11 = self.limit - self.cursor
                        try:
                            if not self.eq_s_b(u"\u0BBE\u0BA9\u0BCD"):
                                raise lab37()
                            v_12 = self.limit - self.cursor
                            v_13 = self.limit - self.cursor
                            try:
                                if not self.eq_s_b(u"\u0B9A"):
                                    raise lab38()
                                raise lab37()
                            except lab38: pass
                            self.cursor = self.limit - v_13
                            self.cursor = self.limit - v_12
                            raise lab36()
                        except lab37: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BBE\u0BB3\u0BCD"):
                                raise lab39()
                            raise lab36()
                        except lab39: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BBE\u0BB0\u0BCD"):
                                raise lab40()
                            raise lab36()
                        except lab40: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BC7\u0BA9\u0BCD"):
                                raise lab41()
                            raise lab36()
                        except lab41: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BBE"):
                                raise lab42()
                            raise lab36()
                        except lab42: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BBE\u0BAE\u0BCD"):
                                raise lab43()
                            raise lab36()
                        except lab43: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BC6\u0BAE\u0BCD"):
                                raise lab44()
                            raise lab36()
                        except lab44: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BC7\u0BAE\u0BCD"):
                                raise lab45()
                            raise lab36()
                        except lab45: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BCB\u0BAE\u0BCD"):
                                raise lab46()
                            raise lab36()
                        except lab46: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0B95\u0BC1\u0BAE\u0BCD"):
                                raise lab47()
                            raise lab36()
                        except lab47: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BA4\u0BC1\u0BAE\u0BCD"):
                                raise lab48()
                            raise lab36()
                        except lab48: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0B9F\u0BC1\u0BAE\u0BCD"):
                                raise lab49()
                            raise lab36()
                        except lab49: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BB1\u0BC1\u0BAE\u0BCD"):
                                raise lab50()
                            raise lab36()
                        except lab50: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BBE\u0BAF\u0BCD"):
                                raise lab51()
                            raise lab36()
                        except lab51: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BC6\u0BA9\u0BCD"):
                                raise lab52()
                            raise lab36()
                        except lab52: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BA9\u0BBF\u0BB0\u0BCD"):
                                raise lab53()
                            raise lab36()
                        except lab53: pass
                        self.cursor = self.limit - v_11
                        try:
                            if not self.eq_s_b(u"\u0BC0\u0BB0\u0BCD"):
                                raise lab54()
                            raise lab36()
                        except lab54: pass
                        self.cursor = self.limit - v_11
                        if not self.eq_s_b(u"\u0BC0\u0BAF\u0BB0\u0BCD"):
                            raise lab35()
                    except lab36: pass
                    self.bra = self.cursor
                    if not self.slice_from(u"\u0BCD"):
                        return False
                    self.B_found_a_match = True
                    self.cursor = self.limit - v_10
                    raise lab1()
                except lab35: pass
                self.cursor = self.limit - v_2
                v_14 = self.limit - self.cursor
                self.ket = self.cursor
                try:
                    v_15 = self.limit - self.cursor
                    try:
                        if not self.eq_s_b(u"\u0B95\u0BC1"):
                            raise lab56()
                        raise lab55()
                    except lab56: pass
                    self.cursor = self.limit - v_15
                    if not self.eq_s_b(u"\u0BA4\u0BC1"):
                        raise lab0()
                except lab55: pass
                v_16 = self.limit - self.cursor
                if not self.eq_s_b(u"\u0BCD"):
                    raise lab0()
                self.cursor = self.limit - v_16
                self.bra = self.cursor
                if not self.slice_del():
                    return False

                self.B_found_a_match = True
                self.cursor = self.limit - v_14
            except lab1: pass
        except lab0: pass
        self.cursor = self.limit - v_1
        v_17 = self.limit - self.cursor
        try:
            self.ket = self.cursor
            if self.find_among_b(TamilStemmer.a_25) == 0:
                raise lab57()
            self.bra = self.cursor
            if not self.slice_del():
                return False

            self.B_found_a_match = True
        except lab57: pass
        self.cursor = self.limit - v_17
        self.cursor = self.limit_backward
        self.__r_fix_endings()
        return True

    def _stem(self):
        self.B_found_vetrumai_urupu = False
        v_1 = self.cursor
        self.__r_fix_ending()
        self.cursor = v_1
        if not self.__r_has_min_length():
            return False
        v_2 = self.cursor
        self.__r_remove_question_prefixes()
        self.cursor = v_2
        v_3 = self.cursor
        self.__r_remove_pronoun_prefixes()
        self.cursor = v_3
        v_4 = self.cursor
        self.__r_remove_question_suffixes()
        self.cursor = v_4
        v_5 = self.cursor
        self.__r_remove_um()
        self.cursor = v_5
        v_6 = self.cursor
        self.__r_remove_common_word_endings()
        self.cursor = v_6
        v_7 = self.cursor
        self.__r_remove_vetrumai_urupukal()
        self.cursor = v_7
        v_8 = self.cursor
        self.__r_remove_plural_suffix()
        self.cursor = v_8
        v_9 = self.cursor
        self.__r_remove_command_suffixes()
        self.cursor = v_9
        v_10 = self.cursor
        self.__r_remove_tense_suffixes()
        self.cursor = v_10
        return True


class lab0(BaseException): pass


class lab1(BaseException): pass


class lab2(BaseException): pass


class lab3(BaseException): pass


class lab4(BaseException): pass


class lab5(BaseException): pass


class lab6(BaseException): pass


class lab7(BaseException): pass


class lab8(BaseException): pass


class lab9(BaseException): pass


class lab10(BaseException): pass


class lab11(BaseException): pass


class lab12(BaseException): pass


class lab13(BaseException): pass


class lab14(BaseException): pass


class lab15(BaseException): pass


class lab16(BaseException): pass


class lab17(BaseException): pass


class lab18(BaseException): pass


class lab19(BaseException): pass


class lab20(BaseException): pass


class lab21(BaseException): pass


class lab22(BaseException): pass


class lab23(BaseException): pass


class lab24(BaseException): pass


class lab25(BaseException): pass


class lab26(BaseException): pass


class lab27(BaseException): pass


class lab28(BaseException): pass


class lab29(BaseException): pass


class lab30(BaseException): pass


class lab31(BaseException): pass


class lab32(BaseException): pass


class lab33(BaseException): pass


class lab34(BaseException): pass


class lab35(BaseException): pass


class lab36(BaseException): pass


class lab37(BaseException): pass


class lab38(BaseException): pass


class lab39(BaseException): pass


class lab40(BaseException): pass


class lab41(BaseException): pass


class lab42(BaseException): pass


class lab43(BaseException): pass


class lab44(BaseException): pass


class lab45(BaseException): pass


class lab46(BaseException): pass


class lab47(BaseException): pass


class lab48(BaseException): pass


class lab49(BaseException): pass


class lab50(BaseException): pass


class lab51(BaseException): pass


class lab52(BaseException): pass


class lab53(BaseException): pass


class lab54(BaseException): pass


class lab55(BaseException): pass


class lab56(BaseException): pass


class lab57(BaseException): pass



from dataclasses import dataclass
from typing import Any, Callable, Generic, List, Mapping, Optional, Sequence, TypeVar

from ..abc import (
    ByteReceiveStream, ByteSendStream, ByteStream, Listener, ObjectReceiveStream, ObjectSendStream,
    ObjectStream, TaskGroup)

T_Item = TypeVar('T_Item')
T_Stream = TypeVar('T_Stream')


@dataclass(eq=False)
class StapledByteStream(ByteStream):
    """
    Combines two byte streams into a single, bidirectional byte stream.

    Extra attributes will be provided from both streams, with the receive stream providing the
    values in case of a conflict.

    :param ByteSendStream send_stream: the sending byte stream
    :param ByteReceiveStream receive_stream: the receiving byte stream
    """

    send_stream: ByteSendStream
    receive_stream: ByteReceiveStream

    async def receive(self, max_bytes: int = 65536) -> bytes:
        return await self.receive_stream.receive(max_bytes)

    async def send(self, item: bytes) -> None:
        await self.send_stream.send(item)

    async def send_eof(self) -> None:
        await self.send_stream.aclose()

    async def aclose(self) -> None:
        await self.send_stream.aclose()
        await self.receive_stream.aclose()

    @property
    def extra_attributes(self) -> Mapping[Any, Callable[[], Any]]:
        return {**self.send_stream.extra_attributes, **self.receive_stream.extra_attributes}


@dataclass(eq=False)
class StapledObjectStream(Generic[T_Item], ObjectStream[T_Item]):
    """
    Combines two object streams into a single, bidirectional object stream.

    Extra attributes will be provided from both streams, with the receive stream providing the
    values in case of a conflict.

    :param ObjectSendStream send_stream: the sending object stream
    :param ObjectReceiveStream receive_stream: the receiving object stream
    """

    send_stream: ObjectSendStream[T_Item]
    receive_stream: ObjectReceiveStream[T_Item]

    async def receive(self) -> T_Item:
        return await self.receive_stream.receive()

    async def send(self, item: T_Item) -> None:
        await self.send_stream.send(item)

    async def send_eof(self) -> None:
        await self.send_stream.aclose()

    async def aclose(self) -> None:
        await self.send_stream.aclose()
        await self.receive_stream.aclose()

    @property
    def extra_attributes(self) -> Mapping[Any, Callable[[], Any]]:
        return {**self.send_stream.extra_attributes, **self.receive_stream.extra_attributes}


@dataclass(eq=False)
class MultiListener(Generic[T_Stream], Listener[T_Stream]):
    """
    Combines multiple listeners into one, serving connections from all of them at once.

    Any MultiListeners in the given collection of listeners will have their listeners moved into
    this one.

    Extra attributes are provided from each listener, with each successive listener overriding any
    conflicting attributes from the previous one.

    :param listeners: listeners to serve
    :type listeners: Sequence[Listener[T_Stream]]
    """

    listeners: Sequence[Listener[T_Stream]]

    def __post_init__(self) -> None:
        listeners: List[Listener[T_Stream]] = []
        for listener in self.listeners:
            if isinstance(listener, MultiListener):
                listeners.extend(listener.listeners)
                del listener.listeners[:]  # type: ignore[attr-defined]
            else:
                listeners.append(listener)

        self.listeners = listeners

    async def serve(self, handler: Callable[[T_Stream], Any],
                    task_group: Optional[TaskGroup] = None) -> None:
        from .. import create_task_group

        async with create_task_group() as tg:
            for listener in self.listeners:
                tg.start_soon(listener.serve, handler, task_group)

    async def aclose(self) -> None:
        for listener in self.listeners:
            await listener.aclose()

    @property
    def extra_attributes(self) -> Mapping[Any, Callable[[], Any]]:
        attributes: dict = {}
        for listener in self.listeners:
            attributes.update(listener.extra_attributes)

        return attributes



import warnings

from ._compat import FlaskWTFDeprecationWarning

warnings.warn(FlaskWTFDeprecationWarning(
    '"flask_wtf.html5" will be removed in 1.0.  '
    'Import directly from "wtforms.fields.html5" '
    'and "wtforms.widgets.html5".'
), stacklevel=2)

from wtforms.widgets.html5 import *
from wtforms.fields.html5 import *



# -*- coding: utf-8 -*-
"""
    livereload.watcher
    ~~~~~~~~~~~~~~~~~~

    A file watch management for LiveReload Server.

    :copyright: (c) 2013 - 2015 by Hsiaoming Yang
    :license: BSD, see LICENSE for more details.
"""

import glob
import logging
import os
import time
import sys

if sys.version_info.major < 3:
    import inspect
else:
    from inspect import signature

try:
    import pyinotify
except ImportError:
    pyinotify = None

logger = logging.getLogger('livereload')


class Watcher(object):
    """A file watcher registry."""
    def __init__(self):
        self._tasks = {}

        # modification time of filepaths for each task,
        # before and after checking for changes
        self._task_mtimes = {}
        self._new_mtimes = {}

        # setting changes
        self._changes = []

        # filepath that is changed
        self.filepath = None
        self._start = time.time()

        # list of ignored dirs
        self.ignored_dirs = ['.git', '.hg', '.svn', '.cvs']

    def ignore_dirs(self, *args):
        self.ignored_dirs.extend(args)

    def remove_dirs_from_ignore(self, *args):
        for a in args:
            self.ignored_dirs.remove(a)

    def ignore(self, filename):
        """Ignore a given filename or not."""
        _, ext = os.path.splitext(filename)
        return ext in ['.pyc', '.pyo', '.o', '.swp']

    def watch(self, path, func=None, delay=0, ignore=None):
        """Add a task to watcher.

        :param path: a filepath or directory path or glob pattern
        :param func: the function to be executed when file changed
        :param delay: Delay sending the reload message. Use 'forever' to
                      not send it. This is useful to compile sass files to
                      css, but reload on changed css files then only.
        :param ignore: A function return True to ignore a certain pattern of
                       filepath.
        """
        self._tasks[path] = {
            'func': func,
            'delay': delay,
            'ignore': ignore,
            'mtimes': {},
        }

    def start(self, callback):
        """Start the watcher running, calling callback when changes are
        observed. If this returns False, regular polling will be used."""
        return False

    def examine(self):
        """Check if there are changes. If so, run the given task.

        Returns a tuple of modified filepath and reload delay.
        """
        if self._changes:
            return self._changes.pop()

        # clean filepath
        self.filepath = None
        delays = set()
        for path in self._tasks:
            item = self._tasks[path]
            self._task_mtimes = item['mtimes']
            changed = self.is_changed(path, item['ignore'])
            if changed:
                func = item['func']
                delay = item['delay']
                if delay and isinstance(delay, float):
                    delays.add(delay)
                if func:
                    name = getattr(func, 'name', None)
                    if not name:
                        name = getattr(func, '__name__', 'anonymous')
                    logger.info(
                        "Running task: {} (delay: {})".format(name, delay))
                    if sys.version_info.major < 3:
                        sig_len = len(inspect.getargspec(func)[0])
                    else:
                        sig_len = len(signature(func).parameters)
                    if sig_len > 0 and isinstance(changed, list):
                        func(changed)
                    else:
                        func()

        if delays:
            delay = max(delays)
        else:
            delay = None
        return self.filepath, delay

    def is_changed(self, path, ignore=None):
        """Check if any filepaths have been added, modified, or removed.

        Updates filepath modification times in self._task_mtimes.
        """
        self._new_mtimes = {}
        changed = False

        if os.path.isfile(path):
            changed = self.is_file_changed(path, ignore)
        elif os.path.isdir(path):
            changed = self.is_folder_changed(path, ignore)
        else:
            changed = self.get_changed_glob_files(path, ignore)

        if not changed:
            changed = self.is_file_removed()

        self._task_mtimes.update(self._new_mtimes)
        return changed

    def is_file_removed(self):
        """Check if any filepaths have been removed since last check.

        Deletes removed paths from self._task_mtimes.
        Sets self.filepath to one of the removed paths.
        """
        removed_paths = set(self._task_mtimes) - set(self._new_mtimes)
        if not removed_paths:
            return False

        for path in removed_paths:
            self._task_mtimes.pop(path)
            # self.filepath seems purely informational, so setting one
            # of several removed files seems sufficient
            self.filepath = path
        return True

    def is_file_changed(self, path, ignore=None):
        """Check if filepath has been added or modified since last check.

        Updates filepath modification times in self._new_mtimes.
        Sets self.filepath to changed path.
        """
        if not os.path.isfile(path):
            return False

        if self.ignore(path):
            return False

        if ignore and ignore(path):
            return False

        mtime = os.path.getmtime(path)

        if path not in self._task_mtimes:
            self._new_mtimes[path] = mtime
            self.filepath = path
            return mtime > self._start

        if self._task_mtimes[path] != mtime:
            self._new_mtimes[path] = mtime
            self.filepath = path
            return True

        self._new_mtimes[path] = mtime
        return False

    def is_folder_changed(self, path, ignore=None):
        """Check if directory path has any changed filepaths."""
        for root, dirs, files in os.walk(path, followlinks=True):
            for d in self.ignored_dirs:
                if d in dirs:
                    dirs.remove(d)

            for f in files:
                if self.is_file_changed(os.path.join(root, f), ignore):
                    return True
        return False

    def get_changed_glob_files(self, path, ignore=None):
        """Check if glob path has any changed filepaths."""
        if sys.version_info[0] >=3 and sys.version_info[1] >=5:
            files = glob.glob(path, recursive=True)
        else:
            files = glob.glob(path)
        changed_files = [f for f in files if self.is_file_changed(f, ignore)]
        return changed_files


class INotifyWatcher(Watcher):
    def __init__(self):
        Watcher.__init__(self)

        self.wm = pyinotify.WatchManager()
        self.notifier = None
        self.callback = None

    def watch(self, path, func=None, delay=None, ignore=None):
        flag = pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_MODIFY
        self.wm.add_watch(path, flag, rec=True, do_glob=True, auto_add=True)
        Watcher.watch(self, path, func, delay, ignore)

    def inotify_event(self, event):
        self.callback()

    def start(self, callback):
        if not self.notifier:
            self.callback = callback

            from tornado import ioloop
            self.notifier = pyinotify.TornadoAsyncNotifier(
                self.wm, ioloop.IOLoop.instance(),
                default_proc_fun=self.inotify_event
            )
            callback()
        return True


def get_watcher_class():
    if pyinotify is None or not hasattr(pyinotify, 'TornadoAsyncNotifier'):
        return Watcher
    return INotifyWatcher



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Base task runner"""
import os
import subprocess
import threading

from airflow.utils.platform import IS_WINDOWS

if not IS_WINDOWS:
    # ignored to avoid flake complaining on Linux
    from pwd import getpwnam  # noqa

from tempfile import NamedTemporaryFile
from typing import Optional, Union

from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from airflow.models.taskinstance import load_error_file
from airflow.utils.configuration import tmp_configuration_copy
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.net import get_hostname
from airflow.utils.platform import getuser

PYTHONPATH_VAR = 'PYTHONPATH'


class BaseTaskRunner(LoggingMixin):
    """
    Runs Airflow task instances by invoking the `airflow tasks run` command with raw
    mode enabled in a subprocess.

    :param local_task_job: The local task job associated with running the
        associated task instance.
    """

    def __init__(self, local_task_job):
        # Pass task instance context into log handlers to setup the logger.
        super().__init__(local_task_job.task_instance)
        self._task_instance = local_task_job.task_instance

        popen_prepend = []
        if self._task_instance.run_as_user:
            self.run_as_user = self._task_instance.run_as_user
        else:
            try:
                self.run_as_user = conf.get('core', 'default_impersonation')
            except AirflowConfigException:
                self.run_as_user = None

        self._error_file = NamedTemporaryFile(delete=True)

        # Add sudo commands to change user if we need to. Needed to handle SubDagOperator
        # case using a SequentialExecutor.
        self.log.debug("Planning to run as the %s user", self.run_as_user)
        if self.run_as_user and (self.run_as_user != getuser()):
            # We want to include any environment variables now, as we won't
            # want to have to specify them in the sudo call - they would show
            # up in `ps` that way! And run commands now, as the other user
            # might not be able to run the cmds to get credentials
            cfg_path = tmp_configuration_copy(chmod=0o600, include_env=True, include_cmds=True)

            # Give ownership of file to user; only they can read and write
            subprocess.check_call(
                ['sudo', 'chown', self.run_as_user, cfg_path, self._error_file.name], close_fds=True
            )

            # propagate PYTHONPATH environment variable
            pythonpath_value = os.environ.get(PYTHONPATH_VAR, '')
            popen_prepend = ['sudo', '-E', '-H', '-u', self.run_as_user]

            if pythonpath_value:
                popen_prepend.append(f'{PYTHONPATH_VAR}={pythonpath_value}')

        else:
            # Always provide a copy of the configuration file settings. Since
            # we are running as the same user, and can pass through environment
            # variables then we don't need to include those in the config copy
            # - the runner can read/execute those values as it needs
            cfg_path = tmp_configuration_copy(chmod=0o600, include_env=False, include_cmds=False)

        self._cfg_path = cfg_path
        self._command = (
            popen_prepend
            + self._task_instance.command_as_list(
                raw=True,
                pickle_id=local_task_job.pickle_id,
                mark_success=local_task_job.mark_success,
                job_id=local_task_job.id,
                pool=local_task_job.pool,
                cfg_path=cfg_path,
            )
            + ["--error-file", self._error_file.name]
        )
        self.process = None

    def deserialize_run_error(self) -> Optional[Union[str, Exception]]:
        """Return task runtime error if its written to provided error file."""
        return load_error_file(self._error_file)

    def _read_task_logs(self, stream):
        while True:
            line = stream.readline()
            if isinstance(line, bytes):
                line = line.decode('utf-8')
            if not line:
                break
            self.log.info(
                'Job %s: Subtask %s %s',
                self._task_instance.job_id,
                self._task_instance.task_id,
                line.rstrip('\n'),
            )

    def run_command(self, run_with=None):
        """
        Run the task command.

        :param run_with: list of tokens to run the task command with e.g. ``['bash', '-c']``
        :return: the process that was run
        :rtype: subprocess.Popen
        """
        run_with = run_with or []
        full_cmd = run_with + self._command

        self.log.info("Running on host: %s", get_hostname())
        self.log.info('Running: %s', full_cmd)

        if IS_WINDOWS:
            proc = subprocess.Popen(
                full_cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                close_fds=True,
                env=os.environ.copy(),
            )
        else:
            proc = subprocess.Popen(
                full_cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                close_fds=True,
                env=os.environ.copy(),
                preexec_fn=os.setsid,
            )

        # Start daemon thread to read subprocess logging output
        log_reader = threading.Thread(
            target=self._read_task_logs,
            args=(proc.stdout,),
        )
        log_reader.daemon = True
        log_reader.start()
        return proc

    def start(self):
        """Start running the task instance in a subprocess."""
        raise NotImplementedError()

    def return_code(self) -> Optional[int]:
        """
        :return: The return code associated with running the task instance or
            None if the task is not yet done.
        :rtype: int
        """
        raise NotImplementedError()

    def terminate(self) -> None:
        """Force kill the running task instance."""
        raise NotImplementedError()

    def on_finish(self) -> None:
        """A callback that should be called when this is done running."""
        if self._cfg_path and os.path.isfile(self._cfg_path):
            if self.run_as_user:
                subprocess.call(['sudo', 'rm', self._cfg_path], close_fds=True)
            else:
                os.remove(self._cfg_path)
        try:
            self._error_file.close()
        except FileNotFoundError:
            # The subprocess has deleted this file before we do
            # so we ignore
            pass



import asyncio
import sys
from typing import Any
from typing import Callable
from typing import Coroutine

import greenlet

from . import compat
from .. import exc


if compat.py37:
    try:
        from contextvars import copy_context as _copy_context

        # If greenlet.gr_context is present in current version of greenlet,
        # it will be set with a copy of the current context on creation.
        # Refs: https://github.com/python-greenlet/greenlet/pull/198
        getattr(greenlet.greenlet, "gr_context")
    except (ImportError, AttributeError):
        _copy_context = None
else:
    _copy_context = None


# implementation based on snaury gist at
# https://gist.github.com/snaury/202bf4f22c41ca34e56297bae5f33fef
# Issue for context: https://github.com/python-greenlet/greenlet/issues/173


class _AsyncIoGreenlet(greenlet.greenlet):
    def __init__(self, fn, driver):
        greenlet.greenlet.__init__(self, fn, driver)
        self.driver = driver
        if _copy_context is not None:
            self.gr_context = _copy_context()


def await_only(awaitable: Coroutine) -> Any:
    """Awaits an async function in a sync method.

    The sync method must be inside a :func:`greenlet_spawn` context.
    :func:`await_` calls cannot be nested.

    :param awaitable: The coroutine to call.

    """
    # this is called in the context greenlet while running fn
    current = greenlet.getcurrent()
    if not isinstance(current, _AsyncIoGreenlet):
        raise exc.MissingGreenlet(
            "greenlet_spawn has not been called; can't call await_() here. "
            "Was IO attempted in an unexpected place?"
        )

    # returns the control to the driver greenlet passing it
    # a coroutine to run. Once the awaitable is done, the driver greenlet
    # switches back to this greenlet with the result of awaitable that is
    # then returned to the caller (or raised as error)
    return current.driver.switch(awaitable)


def await_fallback(awaitable: Coroutine) -> Any:
    """Awaits an async function in a sync method.

    The sync method must be inside a :func:`greenlet_spawn` context.
    :func:`await_` calls cannot be nested.

    :param awaitable: The coroutine to call.

    """
    # this is called in the context greenlet while running fn
    current = greenlet.getcurrent()
    if not isinstance(current, _AsyncIoGreenlet):
        loop = asyncio.get_event_loop()
        if loop.is_running():
            raise exc.MissingGreenlet(
                "greenlet_spawn has not been called and asyncio event "
                "loop is already running; can't call await_() here. "
                "Was IO attempted in an unexpected place?"
            )
        return loop.run_until_complete(awaitable)

    return current.driver.switch(awaitable)


async def greenlet_spawn(
    fn: Callable, *args, _require_await=False, **kwargs
) -> Any:
    """Runs a sync function ``fn`` in a new greenlet.

    The sync function can then use :func:`await_` to wait for async
    functions.

    :param fn: The sync callable to call.
    :param \\*args: Positional arguments to pass to the ``fn`` callable.
    :param \\*\\*kwargs: Keyword arguments to pass to the ``fn`` callable.
    """

    context = _AsyncIoGreenlet(fn, greenlet.getcurrent())
    # runs the function synchronously in gl greenlet. If the execution
    # is interrupted by await_, context is not dead and result is a
    # coroutine to wait. If the context is dead the function has
    # returned, and its result can be returned.
    switch_occurred = False
    try:
        result = context.switch(*args, **kwargs)
        while not context.dead:
            switch_occurred = True
            try:
                # wait for a coroutine from await_ and then return its
                # result back to it.
                value = await result
            except Exception:
                # this allows an exception to be raised within
                # the moderated greenlet so that it can continue
                # its expected flow.
                result = context.throw(*sys.exc_info())
            else:
                result = context.switch(value)
    finally:
        # clean up to avoid cycle resolution by gc
        del context.driver
    if _require_await and not switch_occurred:
        raise exc.AwaitRequired(
            "The current operation required an async execution but none was "
            "detected. This will usually happen when using a non compatible "
            "DBAPI driver. Please ensure that an async DBAPI is used."
        )
    return result


class AsyncAdaptedLock:
    def __init__(self):
        self.mutex = asyncio.Lock()

    def __enter__(self):
        await_fallback(self.mutex.acquire())
        return self

    def __exit__(self, *arg, **kw):
        self.mutex.release()


def _util_async_run_coroutine_function(fn, *args, **kwargs):
    """for test suite/ util only"""

    loop = asyncio.get_event_loop()
    if loop.is_running():
        raise Exception(
            "for async run coroutine we expect that no greenlet or event "
            "loop is running when we start out"
        )
    return loop.run_until_complete(fn(*args, **kwargs))


def _util_async_run(fn, *args, **kwargs):
    """for test suite/ util only"""

    loop = asyncio.get_event_loop()
    if not loop.is_running():
        return loop.run_until_complete(greenlet_spawn(fn, *args, **kwargs))
    else:
        # allow for a wrapped test function to call another
        assert isinstance(greenlet.getcurrent(), _AsyncIoGreenlet)
        return fn(*args, **kwargs)



"""
This module contains a Python interface for Problem Details for HTTP APIs
<https://tools.ietf.org/html/draft-ietf-appsawg-http-problem-00>, which is a standardized format
to communicate distinct "problem types" to non-human consumers.
"""

from .lifecycle import ConnexionResponse


def problem(status, title, detail, type=None, instance=None, headers=None, ext=None):
    """
    Returns a `Problem Details <https://tools.ietf.org/html/draft-ietf-appsawg-http-problem-00>`_ error response.


    :param status: The HTTP status code generated by the origin server for this occurrence of the problem.
    :type status: int
    :param title: A short, human-readable summary of the problem type.  It SHOULD NOT change from occurrence to
                  occurrence of the problem, except for purposes of localisation.
    :type title: str
    :param detail: An human readable explanation specific to this occurrence of the problem.
    :type detail: str
    :param type: An absolute URI that identifies the problem type.  When dereferenced, it SHOULD provide human-readable
                 documentation for the problem type (e.g., using HTML).  When this member is not present its value is
                 assumed to be "about:blank".
    :type: type: str
    :param instance: An absolute URI that identifies the specific occurrence of the problem.  It may or may not yield
                     further information if dereferenced.
    :type instance: str
    :param headers: HTTP headers to include in the response
    :type headers: dict | None
    :param ext: Extension members to include in the body
    :type ext: dict | None
    :return: error response
    :rtype: ConnexionResponse
    """
    if not type:
        type = 'about:blank'

    problem_response = {'type': type, 'title': title, 'detail': detail, 'status': status}
    if instance:
        problem_response['instance'] = instance
    if ext:
        problem_response.update(ext)

    mimetype = content_type = 'application/problem+json'
    return ConnexionResponse(status, mimetype, content_type,
                             body=problem_response,
                             headers=headers)



import functools
import os
import site
import sys
import sysconfig
import typing

from pip._internal.utils import appdirs
from pip._internal.utils.virtualenv import running_under_virtualenv

# Application Directories
USER_CACHE_DIR = appdirs.user_cache_dir("pip")

# FIXME doesn't account for venv linked to global site-packages
site_packages: typing.Optional[str] = sysconfig.get_path("purelib")


def get_major_minor_version() -> str:
    """
    Return the major-minor version of the current Python as a string, e.g.
    "3.7" or "3.10".
    """
    return "{}.{}".format(*sys.version_info)


def get_src_prefix() -> str:
    if running_under_virtualenv():
        src_prefix = os.path.join(sys.prefix, "src")
    else:
        # FIXME: keep src in cwd for now (it is not a temporary folder)
        try:
            src_prefix = os.path.join(os.getcwd(), "src")
        except OSError:
            # In case the current working directory has been renamed or deleted
            sys.exit("The folder you are executing pip from can no longer be found.")

    # under macOS + virtualenv sys.prefix is not properly resolved
    # it is something like /path/to/python/bin/..
    return os.path.abspath(src_prefix)


try:
    # Use getusersitepackages if this is present, as it ensures that the
    # value is initialised properly.
    user_site: typing.Optional[str] = site.getusersitepackages()
except AttributeError:
    user_site = site.USER_SITE


@functools.lru_cache(maxsize=None)
def is_osx_framework() -> bool:
    return bool(sysconfig.get_config_var("PYTHONFRAMEWORK"))



# -*- coding: utf-8 -*-

from flask import Markup, current_app, json
from werkzeug.urls import url_encode

JSONEncoder = json.JSONEncoder

RECAPTCHA_SCRIPT = u'https://www.google.com/recaptcha/api.js'
RECAPTCHA_TEMPLATE = u'''
<script src='%s' async defer></script>
<div class="g-recaptcha" %s></div>
'''

__all__ = ['RecaptchaWidget']


class RecaptchaWidget(object):

    def recaptcha_html(self, public_key):
        html = current_app.config.get('RECAPTCHA_HTML')
        if html:
            return Markup(html)
        params = current_app.config.get('RECAPTCHA_PARAMETERS')
        script = RECAPTCHA_SCRIPT
        if params:
            script += u'?' + url_encode(params)

        attrs = current_app.config.get('RECAPTCHA_DATA_ATTRS', {})
        attrs['sitekey'] = public_key
        snippet = u' '.join([u'data-%s="%s"' % (k, attrs[k]) for k in attrs])
        return Markup(RECAPTCHA_TEMPLATE % (script, snippet))

    def __call__(self, field, error=None, **kwargs):
        """Returns the recaptcha input HTML."""

        try:
            public_key = current_app.config['RECAPTCHA_PUBLIC_KEY']
        except KeyError:
            raise RuntimeError('RECAPTCHA_PUBLIC_KEY config not set')

        return self.recaptcha_html(public_key)



"""Preserve function defaults.

Preserve the default argument values of function signatures in source code
and keep them not evaluated for readability.
"""

import ast
import inspect
import sys
from typing import Any, Dict, List, Optional

from sphinx.application import Sphinx
from sphinx.locale import __
from sphinx.pycode.ast import parse as ast_parse
from sphinx.pycode.ast import unparse as ast_unparse
from sphinx.util import logging

logger = logging.getLogger(__name__)


class DefaultValue:
    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return self.name


def get_function_def(obj: Any) -> ast.FunctionDef:
    """Get FunctionDef object from living object.
    This tries to parse original code for living object and returns
    AST node for given *obj*.
    """
    try:
        source = inspect.getsource(obj)
        if source.startswith((' ', r'\t')):
            # subject is placed inside class or block.  To read its docstring,
            # this adds if-block before the declaration.
            module = ast_parse('if True:\n' + source)
            return module.body[0].body[0]  # type: ignore
        else:
            module = ast_parse(source)
            return module.body[0]  # type: ignore
    except (OSError, TypeError):  # failed to load source code
        return None


def get_default_value(lines: List[str], position: ast.AST) -> Optional[str]:
    try:
        if sys.version_info < (3, 8):  # only for py38+
            return None
        elif position.lineno == position.end_lineno:
            line = lines[position.lineno - 1]
            return line[position.col_offset:position.end_col_offset]
        else:
            # multiline value is not supported now
            return None
    except (AttributeError, IndexError):
        return None


def update_defvalue(app: Sphinx, obj: Any, bound_method: bool) -> None:
    """Update defvalue info of *obj* using type_comments."""
    if not app.config.autodoc_preserve_defaults:
        return

    try:
        lines = inspect.getsource(obj).splitlines()
        if lines[0].startswith((' ', r'\t')):
            lines.insert(0, '')  # insert a dummy line to follow what get_function_def() does.
    except (OSError, TypeError):
        lines = []

    try:
        function = get_function_def(obj)
        if function.args.defaults or function.args.kw_defaults:
            sig = inspect.signature(obj)
            defaults = list(function.args.defaults)
            kw_defaults = list(function.args.kw_defaults)
            parameters = list(sig.parameters.values())
            for i, param in enumerate(parameters):
                if param.default is not param.empty:
                    if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD):
                        default = defaults.pop(0)
                        value = get_default_value(lines, default)
                        if value is None:
                            value = ast_unparse(default)  # type: ignore
                        parameters[i] = param.replace(default=DefaultValue(value))
                    else:
                        default = kw_defaults.pop(0)
                        value = get_default_value(lines, default)
                        if value is None:
                            value = ast_unparse(default)  # type: ignore
                        parameters[i] = param.replace(default=DefaultValue(value))
            sig = sig.replace(parameters=parameters)
            obj.__signature__ = sig
    except (AttributeError, TypeError):
        # failed to update signature (ex. built-in or extension types)
        pass
    except NotImplementedError as exc:  # failed to ast.unparse()
        logger.warning(__("Failed to parse a default argument value for %r: %s"), obj, exc)


def setup(app: Sphinx) -> Dict[str, Any]:
    app.add_config_value('autodoc_preserve_defaults', False, True)
    app.connect('autodoc-before-process-signature', update_defvalue)

    return {
        'version': '1.0',
        'parallel_read_safe': True
    }



from ._api import request, stream
from ._async import (
    AsyncConnectionInterface,
    AsyncConnectionPool,
    AsyncHTTP2Connection,
    AsyncHTTP11Connection,
    AsyncHTTPConnection,
    AsyncHTTPProxy,
    AsyncSOCKSProxy,
)
from ._exceptions import (
    ConnectError,
    ConnectionNotAvailable,
    ConnectTimeout,
    LocalProtocolError,
    NetworkError,
    PoolTimeout,
    ProtocolError,
    ProxyError,
    ReadError,
    ReadTimeout,
    RemoteProtocolError,
    TimeoutException,
    UnsupportedProtocol,
    WriteError,
    WriteTimeout,
)
from ._models import URL, Origin, Request, Response
from ._ssl import default_ssl_context
from ._sync import (
    ConnectionInterface,
    ConnectionPool,
    HTTP2Connection,
    HTTP11Connection,
    HTTPConnection,
    HTTPProxy,
    SOCKSProxy,
)

__all__ = [
    # top-level requests
    "request",
    "stream",
    # models
    "Origin",
    "URL",
    "Request",
    "Response",
    # async
    "AsyncHTTPConnection",
    "AsyncConnectionPool",
    "AsyncHTTPProxy",
    "AsyncHTTP11Connection",
    "AsyncHTTP2Connection",
    "AsyncConnectionInterface",
    "AsyncSOCKSProxy",
    # sync
    "HTTPConnection",
    "ConnectionPool",
    "HTTPProxy",
    "HTTP11Connection",
    "HTTP2Connection",
    "ConnectionInterface",
    "SOCKSProxy",
    # util
    "default_ssl_context",
    # exceptions
    "ConnectionNotAvailable",
    "ProxyError",
    "ProtocolError",
    "LocalProtocolError",
    "RemoteProtocolError",
    "UnsupportedProtocol",
    "TimeoutException",
    "PoolTimeout",
    "ConnectTimeout",
    "ReadTimeout",
    "WriteTimeout",
    "NetworkError",
    "ConnectError",
    "ReadError",
    "WriteError",
]

__version__ = "0.14.7"


__locals = locals()
for __name in __all__:
    if not __name.startswith("__"):
        setattr(__locals[__name], "__module__", "httpcore")  # noqa



import inspect

import six
import sqlalchemy as sa

from .mock import create_mock_engine
from .orm import _get_query_compile_state


def render_expression(expression, bind, stream=None):
    """Generate a SQL expression from the passed python expression.

    Only the global variable, `engine`, is available for use in the
    expression. Additional local variables may be passed in the context
    parameter.

    Note this function is meant for convenience and protected usage. Do NOT
    blindly pass user input to this function as it uses exec.

    :param bind: A SQLAlchemy engine or bind URL.
    :param stream: Render all DDL operations to the stream.
    """

    # Create a stream if not present.

    if stream is None:
        stream = six.moves.cStringIO()

    engine = create_mock_engine(bind, stream)

    # Navigate the stack and find the calling frame that allows the
    # expression to execuate.

    for frame in inspect.stack()[1:]:
        try:
            frame = frame[0]
            local = dict(frame.f_locals)
            local['engine'] = engine
            six.exec_(expression, frame.f_globals, local)
            break
        except Exception:
            pass
    else:
        raise ValueError('Not a valid python expression', engine)

    return stream


def render_statement(statement, bind=None):
    """
    Generate an SQL expression string with bound parameters rendered inline
    for the given SQLAlchemy statement.

    :param statement: SQLAlchemy Query object.
    :param bind:
        Optional SQLAlchemy bind, if None uses the bind of the given query
        object.
    """

    if isinstance(statement, sa.orm.query.Query):
        if bind is None:
            bind = statement.session.get_bind(
                _get_query_compile_state(statement)._mapper_zero()
            )

        statement = statement.statement

    elif bind is None:
        bind = statement.bind

    stream = six.moves.cStringIO()
    engine = create_mock_engine(bind.engine, stream=stream)
    engine.execute(statement)

    return stream.getvalue()



import warnings

from wtforms.ext.csrf.form import SecureForm

warnings.warn(
    "'wtforms.ext.csrf' will be removed in WTForms 3.0. It is built-in"
    " to WTForms core now.",
    DeprecationWarning,
)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import functools
import gzip
import logging
from io import BytesIO as IO
from itertools import chain
from typing import Callable, TypeVar, cast

import pendulum
from flask import after_this_request, g, request
from pendulum.parsing.exceptions import ParserError

from airflow.models import Log
from airflow.utils.session import create_session

T = TypeVar("T", bound=Callable)

logger = logging.getLogger(__name__)


def action_logging(f: T) -> T:
    """Decorator to log user actions"""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        __tracebackhide__ = True  # Hide from pytest traceback.

        with create_session() as session:
            if g.user.is_anonymous:
                user = 'anonymous'
            else:
                user = g.user.username

            fields_skip_logging = {'csrf_token', '_csrf_token'}
            extra_fields = [
                (k, v)
                for k, v in chain(request.values.items(multi=True), request.view_args.items())
                if k not in fields_skip_logging
            ]

            params = {k: v for k, v in chain(request.values.items(), request.view_args.items())}

            log = Log(
                event=f.__name__,
                task_instance=None,
                owner=user,
                extra=str(extra_fields),
                task_id=params.get('task_id'),
                dag_id=params.get('dag_id'),
            )

            if 'execution_date' in request.values:
                execution_date_value = request.values.get('execution_date')
                try:
                    log.execution_date = pendulum.parse(execution_date_value, strict=False)
                except ParserError:
                    logger.exception(
                        "Failed to parse execution_date from the request: %s", execution_date_value
                    )

            session.add(log)

        return f(*args, **kwargs)

    return cast(T, wrapper)


def gzipped(f: T) -> T:
    """Decorator to make a view compressed"""

    @functools.wraps(f)
    def view_func(*args, **kwargs):
        @after_this_request
        def zipper(response):
            accept_encoding = request.headers.get('Accept-Encoding', '')

            if 'gzip' not in accept_encoding.lower():
                return response

            response.direct_passthrough = False

            if (
                response.status_code < 200
                or response.status_code >= 300
                or 'Content-Encoding' in response.headers
            ):
                return response
            gzip_buffer = IO()
            gzip_file = gzip.GzipFile(mode='wb', fileobj=gzip_buffer)
            gzip_file.write(response.data)
            gzip_file.close()

            response.data = gzip_buffer.getvalue()
            response.headers['Content-Encoding'] = 'gzip'
            response.headers['Vary'] = 'Accept-Encoding'
            response.headers['Content-Length'] = len(response.data)

            return response

        return f(*args, **kwargs)

    return cast(T, view_func)
